import pytest

from .. import nodeio
from .. import nodes
from ..nodes import Node


def test_constructor():
    assert Node().data == ()
    assert Node(15).data == '15'
    assert Node('foobar').data == 'foobar'
    assert Node('a', 15, ('b', 13)).data == ('a', '15', ('b', '13'))
    assert Node(()).data == ((), )


def test_eq():
    assert not Node('x') == 'y'
    assert not Node('declare-const', 'x',
                    'Int') == ('declare-const', 'y', 'Int')
    assert not Node('x') == Node('y')
    assert Node('x') == 'x'
    assert Node('declare-const', 'x', 'Int') == ('declare-const', 'x', 'Int')
    assert Node('x') == Node('x')
    assert Node('declare-const', 'x', 'Int') == Node('declare-const', 'x',
                                                     'Int')
    assert Node(Node('pop')) == (Node('pop'), )
    assert Node('pop', 1) == ('pop', 1)

    assert not Node('#b101') == Node('#b100')
    assert not Node('#b101') == 'x'
    assert not Node('#b101') == ('x', )
    assert not Node('#b101') == Node('x')

    assert Node('#b101') == '#b101'
    assert not Node('#b101') == (Node('#b101'), )
    assert Node('#b101') == Node('#b101')
    assert Node('#x1af') == '#x1af'
    assert not Node('#x1af') == (Node('#x1af'), )
    assert Node('#x1af') == Node('#x1af')

    assert not Node('#x1af') == '(_ bv431 12)'
    assert not Node('#b0101') == ('_', 'bv5', 4)
    assert not Node('_', 'bv5', 4) == ('#b0101', )
    assert not Node('_', 'bv431', 12) == ('#x1af', )

    assert not Node('#b110101111') == '#x1af'
    assert not Node('#b000110101111') == ('#x1af', )
    assert not Node('#x1af') == ('#b000110101111', )

    # can't handle any of these cases without BV const normalization
    # each of these denotes equal values
    assert not Node('#b0101') == '(_ bv5 4)'
    assert not Node('#b0101') == Node('_', 'bv5', 4)
    assert not Node('#x1af') == ('_', 'bv431', 12)
    assert not Node('#x1af') == Node('_', 'bv431', 12)
    assert not Node('_', 'bv0', '4') == '#b0000'
    assert not Node(Node('_'), Node('bv5'), Node('4')) == '#b0101'
    assert not Node('_', 'bv5', 4) == Node('#b0101')
    assert not Node('_', 'bv431', 12) == '#x1af'
    assert not Node('_', 'bv431', 12) == Node('#x1af')
    assert not [Node('_', 'bv5', 4)] == ['#b0101']
    assert not Node('#b000110101111') == '#x1af'
    assert not Node('#b000110101111') == Node('#x1af')
    assert not Node('#x1af') == '#b000110101111'
    assert not Node('#x1af') == Node('#b000110101111')


def test_is_leaf():
    assert Node('x').is_leaf()
    assert not Node('not', 'x').is_leaf()


def test_has_ident():
    assert not Node('x').has_ident()
    assert Node('_', 'BitVec', 3).has_ident()
    assert Node('assert', 'x', 'x').has_ident()


def test_get_ident():
    with pytest.raises(AssertionError):
        Node('x').get_ident()
    assert Node('_', 'BitVec', 3).get_ident() == '_'
    assert Node('assert', 'x', 'x').get_ident() == 'assert'


def test_pickle():
    import pickle
    # test __getstate__ and __setstate__ which are used by pickle
    n = Node('not', ('and', 'x', 'y'))
    assert pickle.loads(pickle.dumps(n)) == n


def test_parse_smtlib():
    assert list(nodeio.parse_smtlib('(reset)')) == [Node(Node('reset'))]
    assert list(nodeio.parse_smtlib('(assert (> x y))')) == [
        Node('assert', ('>', 'x', 'y'))
    ]
    input = """
    (set-logic QF_NRA)
    (set-option :source |just for testing|)
    (define-fun x () String "abc""123")
    ; now some other nonsense
    (declare-const y Real
        ; some other comment
    )
    (assert (= (* y y) y))
    (check-sat)
    """
    assert list(nodeio.parse_smtlib(input)) == [
        Node('set-logic', 'QF_NRA'),
        Node('set-option', ':source', '|just for testing|'),
        Node('define-fun', 'x', (), 'String', '"abc""123"'),
        Node('; now some other nonsense\n'),
        Node('declare-const', 'y', 'Real', '; some other comment\n'),
        Node('assert', ('=', ('*', 'y', 'y'), 'y')),
        Node(Node('check-sat')),
    ]

    assert list(
        nodeio.parse_smtlib("(set-option :source |just for testing")) == []
    assert list(nodeio.parse_smtlib("(set-option :source testing")) == []


def test_dfs():
    exprs = [
        Node('assert', ('>', 'x', 'y')),
        Node('assert', ('=', ('*', 'y', 'y'), 'y')),
    ]
    assert list(nodes.dfs(exprs)) == [('assert', ('>', 'x', 'y')), 'assert',
                                      ('>', 'x', 'y'), '>', 'x', 'y',
                                      ('assert', ('=', ('*', 'y', 'y'), 'y')),
                                      'assert', ('=', ('*', 'y', 'y'), 'y'),
                                      '=', ('*', 'y', 'y'), '*', 'y', 'y', 'y']
    assert list(nodes.dfs(exprs, max_depth=2)) == [
        ('assert', ('>', 'x', 'y')), 'assert', ('>', 'x', 'y'),
        ('assert', ('=', ('*', 'y', 'y'), 'y')), 'assert',
        ('=', ('*', 'y', 'y'), 'y')
    ]


def test_substitute():
    x = Node('x')
    expr = Node('assert', ('>', x, 'y'))
    exprs = [expr, Node('assert', ('=', ('*', 'x', 'y'), 'y'))]

    assert nodes.substitute(x, {x: Node('z')}) == Node('z')
    assert nodes.substitute(x, {x.id: Node('z')}) == Node('z')
    assert nodes.substitute(x, {x.id: None}) is None
    assert nodes.substitute(x, {Node('y'): None}) == x

    assert nodes.substitute(expr,
                            {x: Node('z')}) == Node('assert', ('>', 'z', 'y'))
    assert nodes.substitute(expr,
                            {x.id: Node('z')}) == Node('assert',
                                                       ('>', 'z', 'y'))
    assert nodes.substitute(expr, {x.id: None}) == Node('assert', ('>', 'y'))
    assert nodes.substitute(expr, {expr.id: Node('z')}) == Node('z')

    assert nodes.substitute(exprs, {x: Node('z')}) == [
        Node('assert', ('>', 'z', 'y')),
        Node('assert', ('=', ('*', 'z', 'y'), 'y')),
    ]
    assert nodes.substitute(exprs, {x.id: Node('z')}) == [
        Node('assert', ('>', 'z', 'y')),
        Node('assert', ('=', ('*', 'x', 'y'), 'y')),
    ]
    assert nodes.substitute(exprs, {x.id: None}) == [
        Node('assert', ('>', 'y')),
        Node('assert', ('=', ('*', 'x', 'y'), 'y')),
    ]


def test_render_smtlib_expression():
    expr = Node('x')
    assert nodeio.__write_smtlib_str(expr) == 'x'
    assert nodeio.__write_smtlib_pretty_str(expr) == 'x\n'

    expr = Node('; foo')
    assert nodeio.__write_smtlib_str(expr) == '\n; foo\n'
    assert nodeio.__write_smtlib_pretty_str(expr) == '\n; foo\n'

    expr = Node('declare-const', 'x', 'Real')
    assert nodeio.__write_smtlib_str(expr) == '(declare-const x Real)'
    assert nodeio.__write_smtlib_pretty_str(expr) == '(declare-const x Real)\n'

    expr = Node('assert', ('and', ))
    assert nodeio.__write_smtlib_str(expr) == '(assert (and))'
    assert nodeio.__write_smtlib_pretty_str(expr) == '(assert\n  (and)\n)\n'

    expr = Node('assert', ('>', 'x', 'y'), ())
    assert nodeio.__write_smtlib_str(expr) == '(assert (> x y) ())'
    assert nodeio.__write_smtlib_pretty_str(
        expr) == '(assert\n  (> x y)\n  ()\n)\n'


def test_binary_search():
    assert list(nodes.binary_search(4)) == [(2, 4), (0, 2)]
    assert list(nodes.binary_search(15)) == [
        (7, 15),
        (0, 7),
        (11, 15),
        (7, 11),
        (3, 7),
        (0, 3),
    ]



#
# ddSMT: A delta debugger for SMT benchmarks in SMT-Lib v2 format.
#
# This file is part of ddSMT.
#
# Copyright (C) 2013-2021 by the authors listed in AUTHORS file.
#
# ddSMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ddSMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ddSMT.  If not, see <https://www.gnu.org/licenses/>.

import os


def version_from_git(gitdir):
    """Tries to obtain a version string from git based on the last tag."""
    import re
    import subprocess
    try:
        GIT_VERSION = subprocess.check_output(
            ['git', 'describe', '--tags'], cwd=gitdir,
            stderr=subprocess.PIPE).decode('utf8').strip()
    except subprocess.CalledProcessError:
        cnum = subprocess.check_output(['git', 'rev-list', 'HEAD', '--count'],
                                       cwd=gitdir).decode('utf8').strip()
        cid = subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD'],
                                      cwd=gitdir).decode('utf8').strip()
        GIT_VERSION = 'v0.0-{}-g{}'.format(cnum, cid)
    if re.match('^v[0-9.]+$', GIT_VERSION) is not None:
        return GIT_VERSION[1:]
    else:
        m = re.match('v([0-9]+)\\.([0-9]+)-([0-9]+)-g[a-z0-9]+', GIT_VERSION)
        if m is not None:
            return '{}.{}.dev{}'.format(m.group(1),
                                        int(m.group(2)) + 1, m.group(3))
    return GIT_VERSION


def version_from_package_metadata():
    """Tries to obtain a version string from git based on the pypi package
    information."""
    try:
        from importlib import metadata
    except ImportError:
        import importlib_metadata as metadata
    return metadata.version('ddSMT')


__dot_git = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..',
                         '.git')
if os.path.isdir(__dot_git) or os.path.isfile(__dot_git):
    VERSION = version_from_git(os.path.split(__dot_git)[0])
else:
    VERSION = version_from_package_metadata()



#
# ddSMT: A delta debugger for SMT benchmarks in SMT-Lib v2 format.
#
# This file is part of ddSMT.
#
# Copyright (C) 2013-2021 by the authors listed in AUTHORS file.
#
# ddSMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ddSMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ddSMT.  If not, see <https://www.gnu.org/licenses/>.

import collections
import logging
import math
import resource
import subprocess
import sys
import time

from . import nodeio
from . import options
from . import tmpfiles

RunInfo = collections.namedtuple("RunInfo", ["exit", "out", "err", "runtime"])

__GOLDEN = None
__GOLDEN_CC = None


def limit_resources(timeout, pid=None):
    """Apply resource limit given by ``--memout`` and timeout arguments."""
    if pid:
        setlimit = lambda *args: resource.prlimit(pid, *args)  # noqa: E731
    else:
        setlimit = lambda *args: resource.setrlimit(*args)  # noqa: E731

    if options.args().memout:
        setlimit(resource.RLIMIT_AS,
                 (options.args().memout * 1024 * 1024, resource.RLIM_INFINITY))
    if timeout:
        timeout = math.ceil(timeout)
        setlimit(resource.RLIMIT_CPU, (timeout, timeout))


def execute(cmd, filename, timeout):
    """Execute the command on the file with a timeout and a memory limit."""
    if options.args().unchecked:
        return RunInfo(0, "unchecked", "unchecked", 0)
    start = time.time()
    if hasattr(resource, 'prlimit'):
        proc = subprocess.Popen(cmd + [filename],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        limit_resources(timeout, proc.pid)
    else:
        proc = subprocess.Popen(cmd + [filename],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                preexec_fn=lambda: limit_resources(timeout))
    try:
        out, err = proc.communicate(timeout=timeout)
        runtime = time.time() - start
    except subprocess.TimeoutExpired:
        proc.kill()
        logging.debug(f'[!!] timeout: terminated after {timeout:.2f} seconds')
        return RunInfo(proc.returncode, None, None, timeout)
    return RunInfo(proc.returncode, out.decode(), err.decode(), runtime)


def matches_golden(golden, run, ignore_out, match_out, match_err):
    """Checks whether the ``run`` result matches the golden run, considering
    ``ignore_out``, ``match_out`` and ``match_err``.

    If ``ignore_out`` is true, only the exit code is compared.
    Otherwise, if either ``match_out`` or ``match_err`` are given, they
    need to match. Otherwise, both stdout and stderr must be the same.
    """
    if run.exit != golden.exit:
        return False

    if not ignore_out:
        if match_out or match_err:
            # we have --match-out or --match-err
            if match_out:
                if match_out not in run.out:
                    return False
            if match_err:
                if match_err not in run.err:
                    return False
        else:
            if golden.out != run.out:
                return False
            if golden.err != run.err:
                return False

    return True


def check(filename):
    """Check whether the given file behaves as the original input.

    First execute the command and then call ``matches_golden``. If a
    cross-check command is specified, do the same for that one as well.
    """
    ri = execute(options.args().cmd, filename, options.args().timeout)
    if not matches_golden(__GOLDEN, ri,
                          options.args().ignore_output,
                          options.args().match_out,
                          options.args().match_err):
        return False

    if options.args().cmd_cc:
        ri = execute(options.args().cmd_cc, filename,
                     options.args().timeout_cc)
        if not matches_golden(__GOLDEN_CC, ri,
                              options.args().ignore_output_cc,
                              options.args().match_out_cc,
                              options.args().match_err_cc):
            return False

    return True


def check_exprs(exprs):
    """Run the check on the given expressions.

    Returns (True,runtime) if the check was successful and (False,0)
    otherwise.
    """
    tmpfile = tmpfiles.get_tmp_filename()
    nodeio.write_smtlib_for_checking(tmpfile, exprs)
    return check(tmpfile)


def do_golden_runs():
    """Do the initial runs to obtain the golden run results."""
    global __GOLDEN
    global __GOLDEN_CC

    logging.info('')
    if options.args().cmd_cc:
        logging.info('starting initial run, cross checking...')
    else:
        logging.info('starting initial run...')
    logging.info('')

    __GOLDEN = execute(options.args().cmd, options.args().infile, None)

    logging.info(f'golden exit: {__GOLDEN.exit}')
    logging.info(f'golden err:\n{__GOLDEN.err}')
    logging.info(f'golden out:\n{__GOLDEN.out}')
    logging.info(f'golden runtime: {__GOLDEN.runtime:.2f} seconds')
    if options.args().match_out:
        logging.info(f'match (stdout): "{options.args().match_out}"')
        if options.args().match_out not in __GOLDEN.out:
            logging.error(
                f'Expected stdout to match "{options.args().match_out}"')
            sys.exit(1)
    if options.args().match_err:
        logging.info(f'match (stderr): "{options.args().match_err}"')
        if options.args().match_err not in __GOLDEN.err:
            logging.error(
                f'Expected stderr to match "{options.args().match_err}"')
            sys.exit(1)

    if options.args().timeout is None:
        options.args().timeout = round((__GOLDEN.runtime + 1) * 1.5, 2)
        logging.info(
            f'automatic timeout: {options.args().timeout:.2f} seconds')

    if options.args().cmd_cc:
        __GOLDEN_CC = execute(options.args().cmd_cc,
                              options.args().infile, None)

        logging.info("")
        logging.info(f'golden exit (cc): {__GOLDEN_CC.exit}')
        logging.info(f'golden err (cc):\n{__GOLDEN_CC.err}')
        logging.info(f'golden out (cc):\n{__GOLDEN_CC.out}')
        logging.info(f'golden runtime (cc): {__GOLDEN_CC.runtime:.2f} s')
        if options.args().match_out_cc:
            logging.info(
                f'match (cc) (stdout): "{options.args().match_out_cc}"')
        if options.args().match_err_cc:
            logging.info(
                f'match (cc) (stderr): "{options.args().match_err_cc}"')

        if options.args().timeout_cc is None:
            options.args().timeout_cc = round((__GOLDEN_CC.runtime + 1) * 1.5,
                                              2)
            logging.info(
                f'automatic timeout (cc): {options.args().timeout_cc:.2f} s')



from ..nodes import Node
from .. import mutators_core
from .. import options
from .. import smtlib
from .utils import *


def test_binary_reduction():
    m = mutators_core.BinaryReduction()
    assert isinstance(str(m), str)
    assert not hasattr(m, 'filter')
    a = Node('a')
    b = Node('b')
    c = Node('c')
    d = Node('d')
    e = Node('e')
    f = Node('f')
    g = Node('g')
    h = Node('h')
    i = Node('i')
    exprs = [a, b, c, d, e, f, g, h, i]
    mut = [[a, b, c, d], [e, f, g, h, i], [a, b, c, d, e, f],
           [a, b, c, d, g, h, i], [a, b, e, f, g, h, i], [c, d, e, f, g, h, i]]
    assert check_global_mutations(m, b, exprs, [])
    assert check_global_mutations(m, a, exprs, mut)

    assert check_mutations(m, Node('and', 'a', 'b', 'c'), [])
    n = Node('and', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j')
    assert check_mutations(m, n, [
        Node('and', 'a', 'b', 'c', 'd'),
        Node('e', 'f', 'g', 'h', 'i', 'j'),
        Node('and', 'a', 'b', 'c', 'd', 'e', 'f', 'g'),
        Node('and', 'a', 'b', 'c', 'd', 'h', 'i', 'j'),
        Node('and', 'a', 'e', 'f', 'g', 'h', 'i', 'j'),
        Node('b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j'),
    ])


def test_binary_reduction_named():
    m = mutators_core.BinaryReduction()
    m.ident = 'assert'
    assert isinstance(str(m), str)
    assert not hasattr(m, 'filter')
    a = Node('a')
    b = Node('assert', 'b')
    c = Node('c')
    d = Node('assert', 'd')
    e = Node('e')
    f = Node('assert', 'f')
    g = Node('g')
    h = Node('assert', 'h')
    i = Node('assert', 'i')
    exprs = [a, b, c, d, e, f, g, h, i]
    mut = [[a, b, c, d, e, g], [a, c, e, f, g, h, i]]
    assert check_global_mutations(m, a, exprs, mut)


def test_constants():
    smtlib.reset_information()
    m = mutators_core.Constants()
    assert isinstance(str(m), str)
    exprs = [Node('declare-const', 'x', 'Real'), Node('#b1011')]
    assert not m.filter(Node('x'))
    assert list(m.mutations(Node('x'))) == []
    smtlib.collect_information(exprs)
    assert m.filter(Node('x'))
    assert m.filter(Node('#b1011'))
    assert m.filter(Node('=', 'x', '1'))
    assert not m.filter(Node('y'))

    # can't handle this case without BV const normalization
    # else this should be an empty set
    assert check_mutations(
        m, Node('#b0'),
        [Node('_', 'bv0', 1), Node('_', 'bv1', 1)])
    assert check_mutations(m, Node('_', 'bv0', 1), [])

    assert check_mutations(
        m, Node('#b1011'),
        [Node('_', 'bv0', 4), Node('_', 'bv1', 4)])
    assert check_mutations(m, Node('x'), [Node('0.0'), Node('1.0')])
    assert check_mutations(m, Node('=', 'x', '1'),
                           [Node('false'), Node('true')])


def test_erase_node():
    m = mutators_core.EraseNode()
    assert isinstance(str(m), str)
    assert m.filter(Node('x'))
    assert check_mutations(m, Node('x'), [None])


def test_erase_named_node():
    m = mutators_core.EraseNode()
    m.ident = 'assert'
    assert isinstance(str(m), str)
    assert m.filter(Node('assert', 'x'))
    assert not m.filter(Node('x'))
    assert check_mutations(m, Node(Node('assert'), Node('true')), [None])


def test_merge_with_children():
    node = Node(Node('+'), Node('x'), Node(Node('+'), Node('x'), Node('y')))
    m = mutators_core.MergeWithChildren()
    assert isinstance(str(m), str)
    assert m.filter(node)
    assert check_mutations(m, node,
                           [Node(Node('+'), Node('x'), Node('x'), Node('y'))])


def test_replace_by_child():
    smtlib.collect_information([
        Node('declare-const', 'x', 'Int'),
        Node('declare-const', 'y', 'Int'),
    ])
    node = Node('+', 'x', 'y')
    m = mutators_core.ReplaceByChild()
    assert isinstance(str(m), str)
    assert m.filter(node)
    assert not m.filter(Node('x'))
    assert check_mutations(m, node, ['x', 'y'])
    smtlib.reset_information()


def test_replace_by_variable():
    smtlib.reset_information()
    options.args(['in', 'out', 'bin'])

    x = Node('x')
    c = Node('42')
    node = Node(Node('+'), x, c)
    m = mutators_core.ReplaceByVariable()
    assert isinstance(str(m), str)
    assert check_mutations(m, Node('x'), [])

    m.repl_mode = 'inc'
    assert isinstance(str(m), str)
    assert check_mutations(m, node, [])

    v1 = Node('v1')
    v3 = Node('v3')
    exprs = [
        Node(Node('declare-const'), v1, Node('Int')),
        Node(Node('declare-const'), x, Node('Int')),
        Node(Node('declare-const'), Node('v2'), Node('Bool')),
        Node(Node('declare-const'), v3, Node('Int'))
    ]

    smtlib.collect_information(exprs)
    assert m.filter(node)
    # we should not replace the x from the declare-const
    assert not m.filter(x)
    # but we should replace another x from somewhere else
    assert m.filter(Node('x'))
    assert not m.filter(c)
    assert check_mutations(m, node, [v1, x, v3])
    assert check_mutations(m, x, [])
    assert check_mutations(m, v1, [x, v3])
    assert check_mutations(m, v3, [x])

    m.repl_mode = 'dec'
    assert check_mutations(m, node, [v1, x, v3])
    assert check_mutations(m, x, [v1, v3])
    assert check_mutations(m, v1, [])
    assert check_mutations(m, v3, [v1])


def test_sort_children():
    node = Node(Node('+'), Node('123'), Node(Node('*'), Node('y'), Node('2')),
                Node('x'))
    expected = Node(Node('+'), Node('123'), Node('x'),
                    Node(Node('*'), Node('y'), Node('2')))

    m = mutators_core.SortChildren()
    assert isinstance(str(m), str)
    assert not m.filter(Node('x'))
    assert not m.filter(Node('123'))
    assert m.filter(node)
    assert m.filter(expected)
    assert check_mutations(m, node, [expected])
    assert check_mutations(m, expected, [])



import cProfile
import difflib
import glob
import io
import logging
import multiprocessing
import os
import subprocess

from . import nodeio
from . import options


class NodeLoopChecker:
    def __init__(self):
        self.__seen = set()

    def add(self, exprs):
        if options.args().check_loops:
            exprt = tuple(exprs)
            if exprt in self.__seen:
                logging.error(
                    'The following input has already been seen before:')
                logging.error(f'\n{nodeio.write_smtlib_to_str(exprs)}')
                assert False
            self.__seen.add(exprt)


class Profiler:
    """A simple wrapper around cProfile that can be used as context manager,
    honors the ``--profile`` option and automatically writes the results to a
    file. To accumulate all profiles for one process, it uses a static profiler
    and uses a filename pattern that includes the process id.

    Profiling in the main process is somewhat special, as it is enabled
    globally and should thus not be enabled again by a worker function
    that happens to be executed in the main process. Thus, to enable
    profiling in the main process, the Profiler needs to be created with
    ``is_main=True`` to confirm that the caller is aware that it is
    running within the main process. Except for one call in the main
    function, this argument should never be set to True.
    """

    enabled = options.args().profile
    # static profiler, global within one process
    profiler = None

    def __init__(self, is_main=False):
        """Create a profiler."""
        if Profiler.enabled:
            if multiprocessing.parent_process() is None:
                self.enabled = True
                self.filename = '.profile.prof'
            else:
                self.enabled = True
                self.filename = f'.profile-{os.getpid()}.prof'
            if Profiler.profiler is None:
                Profiler.profiler = cProfile.Profile()

    def __enter__(self):
        """Start profiling, if ``--profile`` was given."""
        if Profiler.enabled and self.enabled:
            Profiler.profiler.enable()

    def __exit__(self, type, value, traceback):
        """Stop profiling and write results to file."""
        if Profiler.enabled and self.enabled:
            try:
                Profiler.profiler.disable()
                Profiler.profiler.dump_stats(self.filename)
            except KeyboardInterrupt as e:
                logging.warn(f'Writing {self.filename} was interrupted.'
                             'To avoid corrupted data, we remove this file.')
                os.unlink(self.filename)
                raise e


def __render_profile(infiles, dotfile, pngfile):
    """Helper method to render profile data from a set of input files to a
    png."""
    cmd = [
        'gprof2dot', '--node-label=self-time-percentage',
        '--node-label=total-time', '--node-label=total-time-percentage', '-f',
        'pstats', '-o', dotfile, *infiles
    ]
    try:
        subprocess.run(cmd)
    except FileNotFoundError:
        logging.warn('gprof2dot was not found. Try "pip install gprof2dot".')
        return False

    try:
        subprocess.run(['dot', '-Tpng', '-o', pngfile, dotfile])
    except FileNotFoundError:
        logging.warn('dot was not found. Try "apt install graphviz".')
        return False
    return True


def render_profiles():
    """Convenience function to directly render profiling data to png.

    Splits the data into two parts, the main process and the combination
    of all other processes. Uses ``gprof2dot`` and ``dot``, and will try
    to print helpful error messages if they are not available.
    """
    logging.debug('Rendering profile data')
    if __render_profile(['.profile.prof'], '.profile.dot', 'profile.png'):
        logging.chat('Profile data for main process is in profile.png')

    files = [f for f in glob.glob('.*.prof') if f != '.profile.prof']
    if files:
        if __render_profile(files, '.profile-workder.dot',
                            'profile-worker.png'):
            logging.chat(
                'Profile data for worker processes is in profile-worker.png')


__DIFF_ID = 0


def __render_for_diff(exprs):
    f = io.StringIO()
    for expr in exprs:
        nodeio.__write_smtlib_pretty(f, expr)
    return f.getvalue()


def dump_diff(description, before, after):
    if not options.args().dump_diffs:
        return
    global __DIFF_ID
    b = __render_for_diff(before).splitlines(keepends=True)
    a = __render_for_diff(after).splitlines(keepends=True)

    diff = difflib.unified_diff(b, a)

    __DIFF_ID += 1
    with open(f'.simp-{__DIFF_ID}.diff', 'w') as out:
        out.write(f'{description}\n')
        out.writelines(diff)



from ..nodes import Node
from .. import mutators_arithmetic
from .utils import *


def test_arith_get_mutators():
    d = mutators_arithmetic.get_mutators()
    assert isinstance(d, dict)
    assert len(d) == 4


def test_arith_is_relevant():
    assert mutators_arithmetic.is_relevant(Node('declare-const', 'x', 'Real'))
    assert mutators_arithmetic.is_relevant(Node('declare-fun', 'x', (),
                                                'Real'))
    assert not mutators_arithmetic.is_relevant(
        Node('declare-const', 'x', 'Bool'))
    assert not mutators_arithmetic.is_relevant(
        Node('declare-fun', 'x', (), 'String'))
    assert not mutators_arithmetic.is_relevant(Node())
    assert not mutators_arithmetic.is_relevant(Node('assert', 'x'))


def test_arith_negate_relation():
    m = mutators_arithmetic.ArithmeticNegateRelation()
    assert isinstance(str(m), str)
    notfilter = [
        Node('>', 'x', 'y'),
        Node('=', '1', 'y'),
        Node('distinct', '1', 'y'),
        Node('not', ('and', 'y', 'x')),
    ]
    exprs = {
        Node('not', ('=', 'x', 'y')): [Node('distinct', 'x', 'y')],
        Node('not', ('>', '1', 'y')): [Node('<=', '1', 'y')],
        Node('not', ('distinct', 'x', 'y')): [Node('=', 'x', 'y')],
        Node('not', ('<>', 'x', 'y')): [Node('=', 'x', 'y')],
        Node('not', ('!=', 'x', 'y')): [Node('=', 'x', 'y')],
    }
    for e in notfilter:
        assert not m.filter(e)
    for k, v in exprs.items():
        assert m.filter(k)
        assert check_mutations(m, k, v)


def test_arith_simplify_constant():
    m = mutators_arithmetic.ArithmeticSimplifyConstant()
    assert isinstance(str(m), str)
    notfilter = [
        Node('x'),
        Node('+', '1.0', 'x'),
        Node('0'),
        Node('0.0'),
        Node('1'),
        Node('1.0')
    ]
    exprs = {
        Node('2'): [Node('1'), Node('0')],
        Node('13'): [Node('6'), Node('1')],
        Node('20'): [Node('10'), Node('2')],
        Node('50'): [Node('25'), Node('5')],
        Node('1.5'): [Node('1'), Node('1.')],
        Node('3.1415'): [Node('3'), Node('3.141')],
    }
    for e in notfilter:
        assert not m.filter(e)
    for k, v in exprs.items():
        assert m.filter(k)
        assert check_mutations(m, k, v)

    exprs = [
        Node('assert', ('=', '13', 'y')),
        Node('assert', ('>', '13', 'z')),
        Node('assert', ('<', 'y', 'z')),
    ]
    x = Node('13')
    assert m.filter(x)
    assert check_global_mutations(m, x, exprs, [
        [
            Node('assert', ('=', '6', 'y')),
            Node('assert', ('>', '6', 'z')),
            Node('assert', ('<', 'y', 'z')),
        ],
        [
            Node('assert', ('=', '1', 'y')),
            Node('assert', ('>', '1', 'z')),
            Node('assert', ('<', 'y', 'z')),
        ],
    ])


def test_arith_split_nary_relations():
    m = mutators_arithmetic.ArithmeticSplitNaryRelation()
    assert isinstance(str(m), str)
    notfilter = [Node('>'), Node('>', 'x', 'y'), Node('+', 'x', 'y')]
    exprs = {
        Node('=', 'a', 'b', 'c'):
        [Node('and', ('=', 'a', 'b'), ('=', 'b', 'c'))],
        Node('>', 'a', 'b', 'c'):
        [Node('and', ('>', 'a', 'b'), ('>', 'b', 'c'))],
        Node('<', 'a', 'b', 'c', 'd'):
        [Node('and', ('<', 'a', 'b'), ('<', 'b', 'c'), ('<', 'c', 'd'))],
    }
    for e in notfilter:
        assert not m.filter(e)
    for k, v in exprs.items():
        assert m.filter(k)
        assert check_mutations(m, k, v)


def test_arith_strengthen_relations():
    m = mutators_arithmetic.ArithmeticStrengthenRelation()
    assert isinstance(str(m), str)
    notfilter = [Node('=', 'x', 'y')]
    exprs = {
        Node('<', 'a', 'b'): [Node('=', 'a', 'b')],
        Node('>', 'a', 'b'): [Node('=', 'a', 'b')],
        Node('<=', 'a', 'b'): [Node('<', 'a', 'b'),
                               Node('=', 'a', 'b')],
        Node('>=', 'a', 'b'): [Node('>', 'a', 'b'),
                               Node('=', 'a', 'b')],
        Node('distinct', 'x', 'y'): [Node('=', 'x', 'y')],
    }
    for e in notfilter:
        assert not m.filter(e)
    for k, v in exprs.items():
        assert m.filter(k)
        assert check_mutations(m, k, v)



import pytest
from ..nodes import Node
from .. import smtlib
from ..smtlib import *


def test_collect_information():
    reset_information()
    collect_information([])
    assert smtlib.__constants == {}
    assert smtlib.__defined_functions == {}
    assert smtlib.__sort_lookup == {}

    collect_information([Node('declare-const', 'x')])
    assert smtlib.__constants == {}
    assert smtlib.__defined_functions == {}
    assert smtlib.__sort_lookup == {}

    collect_information([Node('declare-const', ('x', 'Real'), 'Real')])
    assert smtlib.__constants == {}
    assert smtlib.__defined_functions == {}
    assert smtlib.__sort_lookup == {}

    collect_information([Node('declare-const', 'x', 'Real')])
    assert smtlib.__constants == {'x': 'Real'}
    assert smtlib.__defined_functions == {}
    assert smtlib.__sort_lookup == {'x': 'Real'}

    collect_information([Node('declare-fun', 'x', ())])
    assert smtlib.__constants == {}
    assert smtlib.__defined_functions == {}
    assert smtlib.__sort_lookup == {}

    collect_information([Node('declare-fun', ('x', 'Real'), (), 'Real')])
    assert smtlib.__constants == {}
    assert smtlib.__defined_functions == {}
    assert smtlib.__sort_lookup == {}

    collect_information([Node('declare-fun', 'x', 'Real', 'Real')])
    assert smtlib.__constants == {}
    assert smtlib.__defined_functions == {}
    assert smtlib.__sort_lookup == {}

    collect_information([Node('declare-fun', 'x', (), 'Real')])
    assert smtlib.__constants == {'x': 'Real'}
    assert smtlib.__defined_functions == {}
    assert smtlib.__sort_lookup == {'x': 'Real'}

    collect_information([Node('declare-fun', 'x', ('Real', ), 'Real')])
    assert smtlib.__constants == {}
    assert smtlib.__defined_functions == {}
    assert smtlib.__sort_lookup == {'x': 'Real'}

    collect_information([Node('define-fun', 'x', (), 'Real')])
    assert smtlib.__constants == {}
    assert smtlib.__defined_functions == {}
    assert smtlib.__sort_lookup == {}

    collect_information([Node('define-fun', ('x', 'Real'), (), 'Real', '5.0')])
    assert smtlib.__constants == {}
    assert smtlib.__defined_functions == {}
    assert smtlib.__sort_lookup == {}

    collect_information([Node('define-fun', 'x', 'Real', 'Real', '5.0')])
    assert smtlib.__constants == {}
    assert smtlib.__defined_functions == {}
    assert smtlib.__sort_lookup == {}

    collect_information([Node('define-fun', 'x', (), 'Real', '5.0')])
    assert smtlib.__constants == {'x': 'Real'}
    assert 'x' in smtlib.__defined_functions
    assert smtlib.__defined_functions['x'][1](Node()) == '5.0'
    assert smtlib.__sort_lookup == {'x': 'Real'}

    collect_information(
        [Node('define-fun', 'x', ('y', 'Real'), 'Real', ('+', 'y', '5.0'))])
    assert smtlib.__constants == {}
    assert 'x' in smtlib.__defined_functions
    assert smtlib.__defined_functions['x'][1](
        (Node('3.0'), )) == ('+', '3.0', '5.0')
    assert smtlib.__sort_lookup == {'x': 'Real'}


def test_get_variables_with_sort():
    reset_information()
    x = Node('x')
    y = Node('y')
    z = Node('z')
    r = Node('r')
    s = Node('s')
    a = Node('a')
    b = Node('b')
    c = Node('c')
    exprs = [
        Node('declare-const', x, ('_', 'BitVec', 8)),
        Node('declare-fun', y, (), ('_', 'BitVec', 8)),
        Node('declare-const', z, ('_', 'BitVec', 32)),
        Node('declare-fun', r, (), 'Real'),
        Node('declare-const', s, 'String'),
        Node('declare-const', a, 'Float64'),
        Node('declare-const', b, ('_', 'FloatingPoint', 11, 53)),
        Node('declare-fun', c, (), ('_', 'FloatingPoint', 5, 11)),
        Node('bvadd', x, y),
        Node('+', r, r),
        Node('str.++', s, s),
    ]
    assert get_variables_with_sort(Node('_', 'BitVec', 8)) == []
    collect_information(exprs)
    assert get_variables_with_sort(Node('_', 'BitVect', 8)) == []
    assert get_variables_with_sort(Node('_', 'BitVec', 8)) == [x, y]
    assert get_variables_with_sort(Node('_', 'BitVec', 32)) == [z]
    assert get_variables_with_sort(Node('Float64')) == [a]
    assert get_variables_with_sort(Node('_', 'FloatingPoint', 11, 53)) == [b]
    assert get_variables_with_sort(Node('_', 'FloatingPoint', 5, 11)) == [c]
    assert get_variables_with_sort(Node('Real')) == [r]
    assert get_variables_with_sort(Node('String')) == [s]


def test_introduce_variables():
    fp64 = Node('_', 'FloatingPoint', 11, 53)
    x1 = Node('x1')
    x2 = Node('x2')
    x3 = Node('x3')
    decl_x = Node('declare-const', 'x', 'Real')
    decl_x1 = Node('declare-fun', x1, (), fp64)
    decl_x2 = Node('declare-fun', x2, (), fp64)
    decl_x3 = Node('declare-fun', x3, (), fp64)
    fpadd = Node('define-fun', 'fpadd', (), fp64, ('fp.add', 'RTZ', x1, x2))
    fpmul = Node('define-fun', 'fpmul', (), fp64, ('fp.mul', 'RTZ', fpadd, x3))
    fpass = Node('assert', ('=', 'fpmul', 'fpadd'))

    exprs0 = [
        Node('set-info', ':smt-lib-version', '2.6'),
        Node('set-logic', 'QF_FP'),
        Node('set-info', ':category', '"crafted"'),
        Node('set-info', ':status', 'unknown'),
    ]
    assert introduce_variables(exprs0, [decl_x]) \
           == exprs0 + [Node('declare-const', 'x', 'Real')]

    exprs1 = exprs0 + [Node('assert', 'true')]
    assert introduce_variables(exprs1, [decl_x]) \
           == exprs0 + [decl_x, Node('assert', 'true')]

    exprs2 = exprs0 + [decl_x1, decl_x2, decl_x3, fpadd, fpmul, fpass]
    assert introduce_variables(exprs2, [decl_x]) \
           == exprs0 + [decl_x, decl_x1, decl_x2, decl_x3, fpadd, fpmul, fpass]
    exprs3 = [Node('echo')]
    assert introduce_variables(exprs3, [decl_x]) == [decl_x, Node('echo')]


def test_is_leaf():
    assert not is_leaf(Node('declare-const', 'x', 'Real'))
    assert not is_leaf(Node('declare-fun', 'x', (), ('_', 'BitVec', 8)))
    assert not is_leaf(Node('define-fun', 'f', (), 'Int', ('+', 'x', 'x')))
    assert not is_leaf(Node('assert', ('=', 'x', 'x')))
    assert not is_leaf(Node('_', 'BitVec', 2))
    assert is_leaf(Node('x'))
    assert is_leaf(Node('Real'))


def test_is_var():
    reset_information()
    assert not is_var(Node('declare-const', 'x', 'Real'))
    assert not is_var(Node('declare-fun', 'x', (), ('_', 'BitVec', 8)))
    assert not is_var(Node('define-fun', 'f', (), 'Int', ('+', 'x', 'x')))
    assert not is_var(Node('assert', ('=', 'x', 'x')))
    assert not is_var(Node('_', 'BitVec', 2))
    assert not is_var(Node('Real'))
    assert not is_var(Node('x'))
    exprs = [Node('x'), Node('declare-const', 'y', 'Real')]
    collect_information(exprs)
    assert not is_var(Node('x'))
    assert is_var(Node('y'))


def test_is_piped_symbol():
    assert not is_piped_symbol(Node('declare-const', 'x', 'Real'))
    assert not is_piped_symbol(Node('declare-const'))
    assert is_piped_symbol(Node('|declare-const|'))


def test_get_piped_symbol():
    with pytest.raises(AssertionError):
        get_piped_symbol(Node('declare-const', 'x', 'Real'))
    assert get_piped_symbol(Node('|declare-const|')) == 'declare-const'


def test_is_operator_app():
    assert not is_operator_app(Node('bvadd'), 'bvadd')
    assert not is_operator_app(Node('bvad', 'x', 'x'), 'bvadd')
    assert is_operator_app(Node('_', 'BitVec', 3), '_')
    assert is_operator_app(Node('bvadd', 'x', 'x'), 'bvadd')


def test_is_indexed_operator():
    assert not is_indexed_operator(Node('x'), 'extract')
    assert not is_indexed_operator(Node('_', 'sign_extend', 2), 'extract')
    assert not is_indexed_operator(Node('extract', 2), 'extract')
    assert not is_indexed_operator(Node(('_', 'extract', 2), 'x'), 'extract')
    assert not is_indexed_operator(Node('_', 'extract', 2, 2), 'extract')
    assert not is_indexed_operator(Node('_', 'extract', 2), 'extract', 2)
    assert is_indexed_operator(Node('_', 'extract', 2), 'extract')


def test_is_indexed_operator_app():
    assert not is_indexed_operator_app(Node('x'), 'extract')
    assert not is_indexed_operator_app(Node('_', 'sign_extend', 2), 'extract')
    assert not is_indexed_operator_app(Node('extract', 2), 'extract')
    assert not is_indexed_operator_app(Node('_', 'extract', 2, 2), 'extract')
    assert not is_indexed_operator_app(Node('_', 'extract', 2), 'extract', 2)
    assert not is_indexed_operator_app(Node('_', 'extract', 2), 'extract')
    assert is_indexed_operator_app(Node(('_', 'extract', 2), 'x'), 'extract')


def test_has_nary_operator():
    x = Node('x')
    assert not has_nary_operator(x)
    assert has_nary_operator(Node('=>', x, x))
    assert has_nary_operator(Node('and', x, x))
    assert has_nary_operator(Node('or', x, x))
    assert has_nary_operator(Node('xor', x, x))
    assert has_nary_operator(Node('=', x, x))
    assert has_nary_operator(Node('distinct', x, x))
    assert has_nary_operator(Node('+', x, x))
    assert has_nary_operator(Node('-', x, x))
    assert has_nary_operator(Node('*', x, x))
    assert has_nary_operator(Node('div', x, x))
    assert has_nary_operator(Node('/', x, x))
    assert has_nary_operator(Node('<=', x, x))
    assert has_nary_operator(Node('<', x, x))
    assert has_nary_operator(Node('>=', x, x))
    assert has_nary_operator(Node('>', x, x))
    assert has_nary_operator(Node('bvand', x, x))
    assert has_nary_operator(Node('bvor', x, x))
    assert has_nary_operator(Node('bvadd', x, x))
    assert has_nary_operator(Node('bvmul', x, x))
    assert has_nary_operator(Node('concat', x, x))
    assert has_nary_operator(Node('fp.lt', x, x))
    assert has_nary_operator(Node('fp.gt', x, x))
    assert has_nary_operator(Node('fp.leq', x, x))
    assert has_nary_operator(Node('fp.geq', x, x))


def test_is_const():
    assert not is_const(Node('x'))
    assert not is_const(Node('declare-const', 'x', 'Bool'))
    assert not is_const(Node('declare-const', 'x', 'Int'))
    assert not is_const(Node('declare-const', 'x', 'Real'))
    assert not is_const(Node('declare-const', 'x', ('_', 'BitVec', 4)))
    assert not is_const(Node('declare-const', 'x', 'String'))
    assert not is_const(Node('asdf'))
    assert not is_const(Node('1/2'))
    assert not is_const(Node('/', 'x', 2))
    assert not is_const(Node('/', 1, 'x'))
    assert is_const(Node('false'))
    assert is_const(Node('true'))
    assert is_const(Node('/', 1, 2))
    assert is_const(Node('1.0'))
    assert is_const(Node('2'))
    assert is_const(Node('2345'))
    assert is_const(Node('0101'))
    assert is_const(Node('#b0101'))
    assert is_const(Node('#x1af'))
    assert is_const(Node('_', 'bv5', 3))
    assert is_const(Node('"asdf"'))


def test_is_eq():
    assert not is_eq(Node('x'))
    assert not is_eq(Node('>', 'x', 'x'))
    assert is_eq(Node('=', 'x', 'y'))
    assert is_eq(Node('=', 'x', 'y', 'z'))


def test_get_default_constants():
    sort_bool = Node('Bool')
    sort_bv8 = Node('_', 'BitVec', 8)
    sort_bv32 = Node('_', 'BitVec', 32)
    sort_int = Node('Int')
    sort_real = Node('Real')
    sort_fp16 = Node('Float16')
    sort_fp16_1 = Node('_', 'FloatingPoint', 5, 11)
    sort_fp32 = Node('Float32')
    sort_fp32_1 = Node('_', 'FloatingPoint', 8, 24)
    sort_fp64 = Node('Float64')
    sort_fp64_1 = Node('_', 'FloatingPoint', 11, 53)
    sort_fp128 = Node('Float128')
    sort_fp128_1 = Node('_', 'FloatingPoint', 15, 113)
    sort_set = Node('Set', 'Int')
    sort_string = Node('String')
    assert get_default_constants(sort_bool) == [Node('false'), Node('true')]
    assert get_default_constants(sort_bv8) == [
        Node('_', 'bv0', 8), Node('_', 'bv1', 8)
    ]
    assert get_default_constants(sort_bv32) == [
        Node('_', 'bv0', 32), Node('_', 'bv1', 32)
    ]
    assert get_default_constants(sort_int) == [Node(0), Node(1)]
    assert get_default_constants(sort_real) == [Node('0.0'), Node('1.0')]
    assert get_default_constants(sort_fp16) == [
        Node('fp', ('_', 'bv0', 1), ('_', 'bv0', 5), ('_', 'bv0', 10)),
        Node('fp', ('_', 'bv1', 1), ('_', 'bv0', 5), ('_', 'bv0', 10)),
        Node('fp', ('_', 'bv0', 1), ('_', 'bv31', 5), ('_', 'bv1', 10)),
        Node('fp', ('_', 'bv1', 1), ('_', 'bv31', 5), ('_', 'bv1', 10)),
        Node('fp', ('_', 'bv0', 1), ('_', 'bv31', 5), ('_', 'bv0', 10)),
        Node('fp', ('_', 'bv1', 1), ('_', 'bv31', 5), ('_', 'bv0', 10)),
    ]
    assert get_default_constants(sort_fp32) == [
        Node('fp', ('_', 'bv0', 1), ('_', 'bv0', 8), ('_', 'bv0', 23)),
        Node('fp', ('_', 'bv1', 1), ('_', 'bv0', 8), ('_', 'bv0', 23)),
        Node('fp', ('_', 'bv0', 1), ('_', 'bv255', 8), ('_', 'bv1', 23)),
        Node('fp', ('_', 'bv1', 1), ('_', 'bv255', 8), ('_', 'bv1', 23)),
        Node('fp', ('_', 'bv0', 1), ('_', 'bv255', 8), ('_', 'bv0', 23)),
        Node('fp', ('_', 'bv1', 1), ('_', 'bv255', 8), ('_', 'bv0', 23)),
    ]
    assert get_default_constants(sort_fp64) == [
        Node('fp', ('_', 'bv0', 1), ('_', 'bv0', 11), ('_', 'bv0', 52)),
        Node('fp', ('_', 'bv1', 1), ('_', 'bv0', 11), ('_', 'bv0', 52)),
        Node('fp', ('_', 'bv0', 1), ('_', 'bv2047', 11), ('_', 'bv1', 52)),
        Node('fp', ('_', 'bv1', 1), ('_', 'bv2047', 11), ('_', 'bv1', 52)),
        Node('fp', ('_', 'bv0', 1), ('_', 'bv2047', 11), ('_', 'bv0', 52)),
        Node('fp', ('_', 'bv1', 1), ('_', 'bv2047', 11), ('_', 'bv0', 52)),
    ]
    assert get_default_constants(sort_fp128) == [
        Node('fp', ('_', 'bv0', 1), ('_', 'bv0', 15), ('_', 'bv0', 112)),
        Node('fp', ('_', 'bv1', 1), ('_', 'bv0', 15), ('_', 'bv0', 112)),
        Node('fp', ('_', 'bv0', 1), ('_', 'bv32767', 15), ('_', 'bv1', 112)),
        Node('fp', ('_', 'bv1', 1), ('_', 'bv32767', 15), ('_', 'bv1', 112)),
        Node('fp', ('_', 'bv0', 1), ('_', 'bv32767', 15), ('_', 'bv0', 112)),
        Node('fp', ('_', 'bv1', 1), ('_', 'bv32767', 15), ('_', 'bv0', 112)),
    ]
    assert get_default_constants(sort_fp16) == get_default_constants(
        sort_fp16_1)
    assert get_default_constants(sort_fp32) == get_default_constants(
        sort_fp32_1)
    assert get_default_constants(sort_fp64) == get_default_constants(
        sort_fp64_1)
    assert get_default_constants(sort_fp128) == get_default_constants(
        sort_fp128_1)
    assert get_default_constants(sort_set) == [
        Node('as', 'emptyset', ('Set', 'Int')),
        Node('singleton', 0),
        Node('singleton', 1)
    ]
    assert get_default_constants(sort_string) == []


def test_get_sort():
    reset_information()
    bx = Node('bx')
    vx = Node('vx')
    ix = Node('ix')
    rx = Node('rx')
    fx = Node('fx')
    gx = Node('gx')
    rm = Node('rm')
    sx = Node('sx')
    tx = Node('tx')
    sort_bool = Node('Bool')
    sort_bv = Node('_', 'BitVec', 8)
    sort_int = Node('Int')
    sort_real = Node('Real')
    sort_fp16 = Node('_', 'FloatingPoint', 5, 11)
    sort_fp64 = Node('Float64')
    sort_rm = Node('RoundingMode')
    sort_set = Node('Set', 'Int')
    sort_string = Node('String')
    fprem = Node('fp.rem', 'Fx', 'Fx')
    fpfma = Node('fp.fma', rm, fx, fx, fx)

    exprs = [
        Node('declare-const', bx, sort_bool),
        Node('declare-const', vx, sort_bv),
        Node('declare-const', ix, sort_int),
        Node('declare-fun', rx, (), sort_real),
        Node('declare-const', rm, sort_rm),
        Node('define-fun', fx, (Node('x', sort_bool), ), sort_fp16, fprem),
        Node('define-fun', gx, (), sort_fp64, fpfma),
        Node('declare-const', sx, sort_set),
        Node('declare-const', tx, sort_string),
    ]

    assert get_sort(sort_bool) is None
    assert get_sort(sort_bv) is None
    assert get_sort(sort_int) is None
    assert get_sort(sort_real) is None
    assert get_sort(sort_fp16) is None
    assert get_sort(sort_fp64) is None
    assert get_sort(sort_rm) is None
    assert get_sort(sort_set) is None
    assert get_sort(sort_string) is None
    assert get_sort(fpfma) is None

    assert get_sort(Node('true')) == sort_bool
    assert get_sort(Node('false')) == sort_bool
    assert get_sort(Node('#b01011111')) == sort_bv
    assert get_sort(Node('_', 'bv4', 8)) == sort_bv
    assert get_sort(Node(2)) == sort_int
    assert get_sort(Node('2.0')) == sort_real

    assert get_sort(Node('not', bx)) == sort_bool
    assert get_sort(Node('=>', bx, 'true')) == sort_bool
    assert get_sort(Node('and', bx, bx)) == sort_bool
    assert get_sort(Node('or', bx, bx, bx)) == sort_bool
    assert get_sort(Node('xor', bx, bx)) == sort_bool
    assert get_sort(Node('=', fprem, fprem)) == sort_bool
    assert get_sort(Node('distinct', ix, ix)) == sort_bool
    assert get_sort(Node('bvult', vx, vx)) == sort_bool
    assert get_sort(Node('bvule', vx, vx)) == sort_bool
    assert get_sort(Node('bvugt', vx, vx)) == sort_bool
    assert get_sort(Node('bvuge', vx, vx)) == sort_bool
    assert get_sort(Node('bvslt', vx, vx)) == sort_bool
    assert get_sort(Node('bvsle', vx, vx)) == sort_bool
    assert get_sort(Node('bvsgt', vx, vx)) == sort_bool
    assert get_sort(Node('bvsge', vx, vx)) == sort_bool
    assert get_sort(Node('fp.leq', fprem, fprem)) == sort_bool
    assert get_sort(Node('fp.lt', fprem, fprem)) == sort_bool
    assert get_sort(Node('fp.geq', fprem, fprem)) == sort_bool
    assert get_sort(Node('fp.gt', fprem, fprem)) == sort_bool
    assert get_sort(Node('fp.eq', fprem, fprem)) == sort_bool
    assert get_sort(Node('fp.isNormal', fprem)) == sort_bool
    assert get_sort(Node('fp.isSubnormal', fprem)) == sort_bool
    assert get_sort(Node('fp.isZero', fprem)) == sort_bool
    assert get_sort(Node('fp.isInfinite', fprem)) == sort_bool
    assert get_sort(Node('fp.isNaN', fprem)) == sort_bool
    assert get_sort(Node('fp.isNegative', fprem)) == sort_bool
    assert get_sort(Node('fp.isPositive', fprem)) == sort_bool
    assert get_sort(Node('<=', ix, rx)) == sort_bool
    assert get_sort(Node('<', ix, rx)) == sort_bool
    assert get_sort(Node('>>', ix, rx)) == sort_bool
    assert get_sort(Node('>', ix, rx)) == sort_bool
    assert get_sort(Node('is_int', ix)) == sort_bool
    assert get_sort(Node('member', ix, sx)) == sort_bool
    assert get_sort(Node('subset', sx, sx)) == sort_bool
    assert get_sort(Node('str.<', tx, tx)) == sort_bool
    assert get_sort(Node('str.in_re', tx, tx)) == sort_bool
    assert get_sort(Node('str.<=', tx, tx)) == sort_bool
    assert get_sort(Node('str.prefixof', tx, tx)) == sort_bool
    assert get_sort(Node('str.suffixof', tx, tx)) == sort_bool
    assert get_sort(Node('str.contains', tx, tx)) == sort_bool
    assert get_sort(Node('str.is_digit', tx)) == sort_bool
    assert get_sort(Node(('_', 'divisible', 3), ix)) == sort_bool

    assert get_sort(Node('div', ix, ix)) == sort_int
    assert get_sort(Node('mod', ix, ix)) == sort_int
    assert get_sort(Node('abs', ix)) == sort_int
    assert get_sort(Node('to_int', rx)) == sort_int
    assert get_sort(Node('str.len', tx)) == sort_int
    assert get_sort(Node('str.indexof', tx)) == sort_int
    assert get_sort(Node('str.to_code', tx)) == sort_int
    assert get_sort(Node('str.to_int', tx)) == sort_int
    assert get_sort(Node('card', sx)) == sort_int

    assert get_sort(Node('/', rx, rx)) == sort_real
    assert get_sort(Node('+', rx, rx)) is None
    assert get_sort(Node('*', rx, rx)) is None
    assert get_sort(Node('-', rx, rx)) is None
    assert get_sort(Node('to_real', ix)) == sort_real
    assert get_sort(Node('fp.to_real', fpfma)) == sort_real

    assert get_sort(Node('bvcomp', Node('bvcomp', vx, vx), vx)) \
           == Node('_', 'BitVec', 1)

    collect_information(exprs)
    assert get_sort(Node('Fx')) is None
    assert get_sort(bx) == sort_bool
    assert get_sort(vx) == sort_bv
    assert get_sort(ix) == sort_int
    assert get_sort(rx) == sort_real
    assert get_sort(rm) == sort_rm
    assert get_sort(fx) == sort_fp16
    assert get_sort(gx) == sort_fp64
    assert get_sort(sx) == sort_set
    assert get_sort(tx) == sort_string

    assert get_sort(Node('ite', 'true', vx, vx)) == sort_bv
    assert get_sort(Node('ite', 'true', rx, rx)) == sort_real

    assert get_sort(Node('+', ix, ix)) == sort_int
    assert get_sort(Node('*', ix, ix)) == sort_int
    assert get_sort(Node('-', ix, ix)) == sort_int

    assert get_sort(Node('+', rx, rx)) == sort_real
    assert get_sort(Node('*', rx, rx)) == sort_real
    assert get_sort(Node('-', rx, rx)) == sort_real

    assert get_sort(Node('bvadd', Node('bvadd', vx, vx), vx)) == sort_bv
    assert get_sort(Node('bvand', Node('bvand', vx, vx), vx)) == sort_bv
    assert get_sort(Node('bvashr', vx, vx)) == sort_bv
    assert get_sort(Node('bvshr', vx, vx)) == sort_bv
    assert get_sort(Node('bvmul', Node('bvmul', vx, vx), vx)) == sort_bv
    assert get_sort(Node('bvnand', vx, vx)) == sort_bv
    assert get_sort(Node('bvneg', Node('bvnand', vx, vx))) == sort_bv
    assert get_sort(Node('bvnor', vx, vx)) == sort_bv
    assert get_sort(Node('bvnot', Node('bvnand', vx, vx))) == sort_bv
    assert get_sort(Node('bvor', vx, vx)) == sort_bv
    assert get_sort(Node('bvsdiv', Node('bvand', vx, vx), vx)) == sort_bv
    assert get_sort(Node('bvshl', vx, vx)) == sort_bv
    assert get_sort(Node('bvsmod', Node('bvshr', vx, vx), vx)) == sort_bv
    assert get_sort(Node('bvsrem', vx, vx)) == sort_bv
    assert get_sort(Node('bvsub', Node('bvmul', vx, vx), vx)) == sort_bv
    assert get_sort(Node('bvudiv', vx, vx)) == sort_bv
    assert get_sort(Node('bvurem', Node('bvmul', vx, vx), vx)) == sort_bv
    assert get_sort(Node('bvxnor', vx, vx))
    assert get_sort(Node('bvxor', Node('bvshr', vx, vx), vx)) == sort_bv

    assert get_sort(Node('concat', vx, vx)) == Node('_', 'BitVec', 16)
    assert get_sort(Node('concat', Node('concat', vx, vx), vx)) \
           == Node('_', 'BitVec', 24)

    assert get_sort(Node('fp.abs', fx)) == sort_fp16
    assert get_sort(Node('fp.max', gx, gx)) == sort_fp64
    assert get_sort(Node('fp.min', gx, gx)) == sort_fp64
    assert get_sort(Node('fp.neg', fx)) == sort_fp16
    assert get_sort(Node('fp.rem', fx, fx)) == sort_fp16
    assert get_sort(Node('fp.add', rm, gx, gx)) == sort_fp64
    assert get_sort(Node('fp.div', rm, gx, gx)) == sort_fp64
    assert get_sort(Node('fp.fma', rm, gx, gx)) == sort_fp64
    assert get_sort(Node('fp.mul', rm, gx, gx, gx)) == sort_fp64
    assert get_sort(Node('fp.roundToIntegral', rm, fx)) == sort_fp16
    assert get_sort(Node('fp.sqrt', rm, fx)) == sort_fp16
    assert get_sort(Node('fp.sub', rm, fx, fx)) == sort_fp16
    assert get_sort(Node('fp', '#b1', '#b11011', '#b1011011010')) == sort_fp16

    assert get_sort(Node(('_', 'extract', 3, 1), vx)) == Node('_', 'BitVec', 3)
    assert get_sort(Node(('_', 'repeat', 3), vx)) == Node('_', 'BitVec', 24)
    assert get_sort(Node(('_', 'rotate_left', 3), vx)) == sort_bv
    assert get_sort(Node(('_', 'rotate_right', 3), vx)) == sort_bv
    assert get_sort(Node(('_', 'sign_extend', 3), vx)) \
           == Node('_', 'BitVec', 11)
    assert get_sort(Node(('_', 'zero_extend', 3), vx)) \
           == Node('_', 'BitVec', 11)

    assert get_sort(Node(('_', 'fp.to_sbv', 8), fx)) == sort_bv
    assert get_sort(Node(('_', 'fp.to_ubv', 8), fx)) == sort_bv

    assert get_sort(Node(('_', 'to_fp', 5, 11), rm, rx)) == sort_fp16
    assert get_sort(Node(('_', 'to_fp', 5, 11), rm, fx)) == sort_fp16
    assert get_sort(Node(('_', 'to_fp', 5, 11), vx)) == sort_fp16
    assert get_sort(Node(('_', 'to_fp', 5, 11), rm, vx)) == sort_fp16
    assert get_sort(Node(('_', 'to_fp_unsigned', 5, 11), rm, vx)) == sort_fp16


def test_get_indices():
    with pytest.raises(AssertionError):
        get_indices(Node('x'), 'x')
    with pytest.raises(ValueError):
        get_indices(Node('_', 'zero_extend', 'asdf'), 'zero_extend')
    assert get_indices(Node('_', 'zero_extend', 2), 'zero_extend') == [2]
    assert get_indices(Node('_', 'extract', 2, 1), 'extract', 2) == [2, 1]


def test_get_bv_constant_value():
    with pytest.raises(AssertionError):
        get_bv_constant_value(Node('x'))
    with pytest.raises(AssertionError):
        get_bv_constant_value(Node('#c3'))
    assert get_bv_constant_value(Node('_', 'bv3', 3)) == (3, 3)
    assert get_bv_constant_value(Node('#b011')) == (3, 3)
    assert get_bv_constant_value(Node('#x3')) == (3, 4)


def test_is_fp_sort():
    assert not is_fp_sort(Node('x'))
    assert not is_fp_sort(Node('Real'))
    assert not is_fp_sort(Node('_', 'BitVec', 3))
    assert not is_fp_sort(Node('_', 'FloatingPoint', 3))
    assert not is_fp_sort(Node('_', 'FloatinPoint', 3, 5))
    assert is_fp_sort(Node('Float16'))
    assert is_fp_sort(Node('Float32'))
    assert is_fp_sort(Node('Float64'))
    assert is_fp_sort(Node('Float128'))
    assert is_fp_sort(Node('_', 'FloatingPoint', 3, 5))


def test_is_bool_const():
    assert not is_bool_const(Node('x'))
    assert not is_bool_const(Node('declare-const', 'x', 'Bool'))
    assert is_bool_const(Node('false'))
    assert is_bool_const(Node('true'))


def test_is_arith_const():
    assert not is_arith_const(Node('x'))
    assert not is_arith_const(Node('declare-const', 'x', 'Int'))
    assert not is_arith_const(Node('declare-const', 'x', 'Real'))
    assert not is_arith_const(Node('1/2'))
    assert not is_const(Node('/', 'x', 2))
    assert not is_const(Node('/', 1, 'x'))
    assert is_arith_const(Node('/', 1, 2))
    assert is_arith_const(Node('1.0'))
    assert is_arith_const(Node('2'))
    assert is_arith_const(Node('2345'))


def test_is_int_const():
    assert not is_int_const(Node('x'))
    assert not is_int_const(Node('declare-const', 'x', 'Int'))
    assert not is_int_const(Node('1.0'))
    assert not is_int_const(Node('1/2'))
    assert is_int_const(Node('2'))
    assert is_int_const(Node('2345'))


def test_is_real_const():
    assert not is_real_const(Node('x'))
    assert not is_real_const(Node('declare-const', 'x', 'Real'))
    assert not is_real_const(Node('1/2'))
    assert not is_const(Node('/', 'x', 2))
    assert not is_const(Node('/', 1, 'x'))
    assert is_real_const(Node('/', 1, 2))
    assert is_real_const(Node('2'))
    assert is_real_const(Node('2345'))
    assert is_real_const(Node('1.0'))


def test_is_bv_const():
    assert not is_bv_const(Node('x'))
    assert not is_bv_const(Node('declare-const', 'x', ('_', 'BitVec', 4)))
    assert not is_bv_const(Node('2'))
    assert not is_bv_const(Node('0101'))
    assert not is_bv_const(Node('_', ('bv5', ), 3))
    assert not is_bv_const(Node('#b0121'))
    assert not is_bv_const(Node('#x1az'))
    assert is_bv_const(Node('#b0101'))
    assert is_bv_const(Node('#x1af'))
    assert is_bv_const(Node('#x1AF'))
    assert is_bv_const(Node('_', 'bv5', 3))


def test_is_string_const():
    assert not is_string_const(Node('x'))
    assert not is_string_const(Node('declare-const', 'x', 'String'))
    assert not is_string_const(Node('2'))
    assert not is_string_const(Node('asdf'))
    assert is_string_const(Node('"asdf"'))


def test_is_bv_sort():
    assert not is_bv_sort(Node('declare-const', 'x', ('_', 'BitVec', 8)))
    assert not is_bv_sort(Node('_', 'BitVec'))
    assert not is_bv_sort(Node('BitVec', 8))
    assert not is_bv_sort(Node('_', 'BitVe', 8))
    assert is_bv_sort(Node('_', 'BitVec', 8))


def test_is_bv_comp():
    assert not is_bv_comp(
        Node('define-fun', 'x', (), ('_', 'BitVec', 8), ('bvcomp', 'x', 'x')))
    assert not is_bv_comp(Node('=', 'x', 'x'))
    assert is_bv_comp(Node('bvcomp', 'x', 'x'))


def test_is_bv_not():
    assert not is_bv_not(
        Node('define-fun', 'x', (), ('_', 'BitVec', 8), ('bvnot', 'x', 'x')))
    assert not is_bv_not(Node('not', 'x', 'x'))
    assert is_bv_not(Node('bvnot', 'x', 'x'))


def test_is_bv_neg():
    assert not is_bv_neg(
        Node('define-fun', 'x', (), ('_', 'BitVec', 8), ('bvneg', 'x', 'x')))
    assert not is_bv_neg(Node('-', 'x', 'x'))
    assert is_bv_neg(Node('bvneg', 'x', 'x'))


def test_get_bv_width():
    reset_information()
    bx = Node('bx')
    vx = Node('vx')
    ix = Node('ix')
    rx = Node('rx')
    fx = Node('fx')
    gx = Node('gx')
    rm = Node('rm')
    sx = Node('sx')
    tx = Node('tx')
    sort_bool = Node('Bool')
    sort_bv = Node('_', 'BitVec', 8)
    sort_int = Node('Int')
    sort_real = Node('Real')
    sort_fp16 = Node('_', 'FloatingPoint', 5, 11)
    sort_fp64 = Node('Float64')
    sort_rm = Node('RoundingMode')
    sort_set = Node('Set', 'Int')
    sort_string = Node('String')
    fprem = Node('fp.rem', 'Fx', 'Fx')
    fpfma = Node('fp.fma', rm, fx, fx, fx)

    exprs = [
        Node('declare-const', bx, sort_bool),
        Node('declare-const', vx, sort_bv),
        Node('declare-const', ix, sort_int),
        Node('declare-fun', rx, (), sort_real),
        Node('declare-const', rm, sort_rm),
        Node('define-fun', fx, (Node('x', sort_bool), ), sort_fp16, fprem),
        Node('define-fun', gx, (), sort_fp64, fpfma),
        Node('declare-const', sx, sort_set),
        Node('declare-const', tx, sort_string),
    ]

    assert get_bv_width(sort_bool) == -1
    assert get_bv_width(sort_bv) == -1
    assert get_bv_width(sort_int) == -1
    assert get_bv_width(sort_real) == -1
    assert get_bv_width(sort_fp16) == -1
    assert get_bv_width(sort_fp64) == -1
    assert get_bv_width(sort_rm) == -1
    assert get_bv_width(sort_set) == -1
    assert get_bv_width(sort_string) == -1
    assert get_bv_width(fpfma) == -1

    assert get_bv_width(Node('true')) == -1
    assert get_bv_width(Node('false')) == -1
    assert get_bv_width(Node('#b01011111')) == 8
    assert get_bv_width(Node('#x110')) == 12
    assert get_bv_width(Node('_', 'bv4', 8)) == 8
    assert get_bv_width(Node(2)) == -1
    assert get_bv_width(Node('2.0')) == -1

    assert get_bv_width(Node('not', bx)) == -1
    assert get_bv_width(Node('=>', bx, 'true')) == -1
    assert get_bv_width(Node('and', bx, bx)) == -1
    assert get_bv_width(Node('or', bx, bx, bx)) == -1
    assert get_bv_width(Node('xor', bx, bx)) == -1
    assert get_bv_width(Node('=', fprem, fprem)) == -1
    assert get_bv_width(Node('distinct', ix, ix)) == -1
    assert get_bv_width(Node('bvult', vx, vx)) == -1
    assert get_bv_width(Node('bvule', vx, vx)) == -1
    assert get_bv_width(Node('bvugt', vx, vx)) == -1
    assert get_bv_width(Node('bvuge', vx, vx)) == -1
    assert get_bv_width(Node('bvslt', vx, vx)) == -1
    assert get_bv_width(Node('bvsle', vx, vx)) == -1
    assert get_bv_width(Node('bvsgt', vx, vx)) == -1
    assert get_bv_width(Node('bvsge', vx, vx)) == -1
    assert get_bv_width(Node('fp.leq', fprem, fprem)) == -1
    assert get_bv_width(Node('fp.lt', fprem, fprem)) == -1
    assert get_bv_width(Node('fp.geq', fprem, fprem)) == -1
    assert get_bv_width(Node('fp.gt', fprem, fprem)) == -1
    assert get_bv_width(Node('fp.eq', fprem, fprem)) == -1
    assert get_bv_width(Node('fp.isNormal', fprem)) == -1
    assert get_bv_width(Node('fp.isSubnormal', fprem)) == -1
    assert get_bv_width(Node('fp.isZero', fprem)) == -1
    assert get_bv_width(Node('fp.isInfinite', fprem)) == -1
    assert get_bv_width(Node('fp.isNaN', fprem)) == -1
    assert get_bv_width(Node('fp.isNegative', fprem)) == -1
    assert get_bv_width(Node('fp.isPositive', fprem)) == -1
    assert get_bv_width(Node('<=', ix, rx)) == -1
    assert get_bv_width(Node('<', ix, rx)) == -1
    assert get_bv_width(Node('>>', ix, rx)) == -1
    assert get_bv_width(Node('>', ix, rx)) == -1
    assert get_bv_width(Node('is_int', ix)) == -1
    assert get_bv_width(Node('member', ix, sx)) == -1
    assert get_bv_width(Node('subset', sx, sx)) == -1
    assert get_bv_width(Node('str.<', tx, tx)) == -1
    assert get_bv_width(Node('str.in_re', tx, tx)) == -1
    assert get_bv_width(Node('str.<=', tx, tx)) == -1
    assert get_bv_width(Node('str.prefixof', tx, tx)) == -1
    assert get_bv_width(Node('str.suffixof', tx, tx)) == -1
    assert get_bv_width(Node('str.contains', tx, tx)) == -1
    assert get_bv_width(Node('str.is_digit', tx)) == -1
    assert get_bv_width(Node(('_', 'divisible', 3), ix)) == -1

    assert get_bv_width(Node('div', ix, ix)) == -1
    assert get_bv_width(Node('mod', ix, ix)) == -1
    assert get_bv_width(Node('abs', ix)) == -1
    assert get_bv_width(Node('to_int', rx)) == -1
    assert get_bv_width(Node('str.len', tx)) == -1
    assert get_bv_width(Node('str.indexof', tx)) == -1
    assert get_bv_width(Node('str.to_code', tx)) == -1
    assert get_bv_width(Node('str.to_int', tx)) == -1
    assert get_bv_width(Node('card', sx)) == -1

    assert get_bv_width(Node('/', rx, rx)) == -1
    assert get_bv_width(Node('+', rx, rx)) == -1
    assert get_bv_width(Node('*', rx, rx)) == -1
    assert get_bv_width(Node('-', rx, rx)) == -1
    assert get_bv_width(Node('to_real', ix)) == -1
    assert get_bv_width(Node('fp.to_real', fpfma)) == -1

    assert get_bv_width(Node('bvcomp', Node('bvcomp', vx, vx), vx)) == 1

    collect_information(exprs)
    assert get_bv_width(Node('Fx')) == -1
    assert get_bv_width(bx) == -1
    assert get_bv_width(vx) == 8
    assert get_bv_width(ix) == -1
    assert get_bv_width(rx) == -1
    assert get_bv_width(rm) == -1
    assert get_bv_width(fx) == -1
    assert get_bv_width(gx) == -1
    assert get_bv_width(sx) == -1
    assert get_bv_width(tx) == -1

    assert get_bv_width(Node('ite', 'true', vx, vx)) == 8

    assert get_bv_width(Node('+', ix, ix)) == -1
    assert get_bv_width(Node('*', ix, ix)) == -1
    assert get_bv_width(Node('-', ix, ix)) == -1

    assert get_bv_width(Node('+', rx, rx)) == -1
    assert get_bv_width(Node('*', rx, rx)) == -1
    assert get_bv_width(Node('-', rx, rx)) == -1

    assert get_bv_width(Node('bvadd', Node('bvadd', vx, vx), vx)) == 8
    assert get_bv_width(Node('bvand', Node('bvand', vx, vx), vx)) == 8
    assert get_bv_width(Node('bvashr', vx, vx)) == 8
    assert get_bv_width(Node('bvshr', vx, vx)) == 8
    assert get_bv_width(Node('bvmul', Node('bvmul', vx, vx), vx)) == 8
    assert get_bv_width(Node('bvnand', vx, vx)) == 8
    assert get_bv_width(Node('bvneg', Node('bvnand', vx, vx))) == 8
    assert get_bv_width(Node('bvnor', vx, vx)) == 8
    assert get_bv_width(Node('bvnot', Node('bvnand', vx, vx))) == 8
    assert get_bv_width(Node('bvor', vx, vx)) == 8
    assert get_bv_width(Node('bvsdiv', Node('bvand', vx, vx), vx)) == 8
    assert get_bv_width(Node('bvshl', vx, vx)) == 8
    assert get_bv_width(Node('bvsmod', Node('bvshr', vx, vx), vx)) == 8
    assert get_bv_width(Node('bvsrem', vx, vx)) == 8
    assert get_bv_width(Node('bvsub', Node('bvmul', vx, vx), vx)) == 8
    assert get_bv_width(Node('bvudiv', vx, vx)) == 8
    assert get_bv_width(Node('bvurem', Node('bvmul', vx, vx), vx)) == 8
    assert get_bv_width(Node('bvxnor', vx, vx))
    assert get_bv_width(Node('bvxor', Node('bvshr', vx, vx), vx)) == 8

    assert get_bv_width(Node('concat', vx, vx)) == 16
    assert get_bv_width(Node('concat', Node('concat', vx, vx), vx)) == 24

    assert get_bv_width(Node('fp.abs', fx)) == -1
    assert get_bv_width(Node('fp.max', gx, gx)) == -1
    assert get_bv_width(Node('fp.min', gx, gx)) == -1
    assert get_bv_width(Node('fp.neg', fx)) == -1
    assert get_bv_width(Node('fp.rem', fx, fx)) == -1
    assert get_bv_width(Node('fp.add', rm, gx, gx)) == -1
    assert get_bv_width(Node('fp.div', rm, gx, gx)) == -1
    assert get_bv_width(Node('fp.fma', rm, gx, gx)) == -1
    assert get_bv_width(Node('fp.mul', rm, gx, gx, gx)) == -1
    assert get_bv_width(Node('fp.roundToIntegral', rm, fx)) == -1
    assert get_bv_width(Node('fp.sqrt', rm, fx)) == -1
    assert get_bv_width(Node('fp.sub', rm, fx, fx)) == -1
    assert get_bv_width(Node('fp', '#b1', '#b11011', '#b1011011010')) == -1

    assert get_bv_width(Node(('_', 'extract', 3, 1), vx)) == 3
    assert get_bv_width(Node(('_', 'repeat', 3), vx)) == 24
    assert get_bv_width(Node(('_', 'rotate_left', 3), vx)) == 8
    assert get_bv_width(Node(('_', 'rotate_right', 3), vx)) == 8
    assert get_bv_width(Node(('_', 'sign_extend', 3), vx)) == 11
    assert get_bv_width(Node(('_', 'zero_extend', 3), vx)) == 11

    assert get_bv_width(Node(('_', 'fp.to_sbv', 8), fx)) == 8
    assert get_bv_width(Node(('_', 'fp.to_ubv', 8), fx)) == 8

    assert get_bv_width(Node(('_', 'to_fp', 5, 11), rm, rx)) == -1
    assert get_bv_width(Node(('_', 'to_fp', 5, 11), rm, fx)) == -1
    assert get_bv_width(Node(('_', 'to_fp', 5, 11), vx)) == -1
    assert get_bv_width(Node(('_', 'to_fp', 5, 11), rm, vx)) == -1
    assert get_bv_width(Node(('_', 'to_fp_unsigned', 5, 11), rm, vx)) == -1


def test_is_defined_fun():
    reset_information()
    x1 = Node('x1')
    x2 = Node('x2')
    x3 = Node('x3')
    fp64 = Node('_', 'FloatingPoint', 11, 53)
    fpadd = Node('fp.add', 'RTZ', x1, x2)
    fpmul = Node('fp.mul', 'RTZ', fpadd, x3)
    fprem = Node('fp.rem', 'x', 'x')
    fadd = Node('define-fun', 'fadd', (), fp64, fpadd)
    fmul = Node('define-fun', 'fmul', (), fp64, fpmul)
    fx = Node('define-fun', 'fx', (Node('x', 'Bool'), ), fp64, fprem)
    fpfma = Node('fp.fma', 'RNE', 'fx', 'fx', 'fx')
    gx = Node('define-fun', 'gx', (
        Node('x', 'Bool'),
        Node('y', 'Int'),
    ), fp64, fpfma)
    f = Node('define-fun', 'f', (Node('x', 'Int'), ), 'Int', ('+', 'x', 'x'))
    exprs = [
        fadd,
        fmul,
        fx,
        gx,
        f,
        Node('declare-const', 'x', 'Bool'),
        Node('declare-fun', 'y', 'Int'),
    ]
    collect_information(exprs)
    assert not is_defined_fun(x1)
    assert not is_defined_fun(x2)
    assert not is_defined_fun(x3)
    assert not is_defined_fun(fpadd)
    assert not is_defined_fun(fpmul)
    assert not is_defined_fun(fprem)
    assert not is_defined_fun(fpfma)
    assert not is_defined_fun(fadd)
    assert not is_defined_fun(fmul)
    assert not is_defined_fun(fx)
    assert not is_defined_fun(gx)
    assert not is_defined_fun(f)
    assert is_defined_fun(Node('fadd'))
    assert is_defined_fun(Node('fmul'))
    assert is_defined_fun(Node('fx'))
    assert is_defined_fun(Node('gx'))
    assert is_defined_fun(Node('f'))
    assert is_defined_fun(Node('fx', 'true'))
    assert is_defined_fun(Node('gx', 'true', 2))
    assert is_defined_fun(Node('f', 15))


def test_get_defined_fun():
    reset_information()
    x1 = Node('x1')
    x2 = Node('x2')
    x3 = Node('x3')
    fp64 = Node('_', 'FloatingPoint', 11, 53)
    fpadd = Node('fp.add', 'RTZ', x1, x2)
    fpmul = Node('fp.mul', 'RTZ', fpadd, x3)
    fprem = Node('fp.rem', 'x', 'x')
    fadd = Node('define-fun', 'fadd', (), fp64, fpadd)
    fmul = Node('define-fun', 'fmul', (), fp64, fpmul)
    fx = Node('define-fun', 'fx', (Node('x', 'Bool'), ), fp64, fprem)
    fpfma = Node('fp.fma', 'y', 'fx', 'fx', 'fx')
    gx = Node('define-fun', 'gx', (
        Node('x', 'Bool'),
        Node('y', 'Int'),
    ), fp64, fpfma)
    f = Node('define-fun', 'f', (Node('x', 'Int'), ), 'Int', ('+', 'x', 'x'))
    exprs = [
        fadd,
        fmul,
        fx,
        gx,
        f,
        Node('declare-const', 'x', 'Bool'),
        Node('declare-fun', 'y', 'Int'),
    ]
    collect_information(exprs)
    with pytest.raises(AssertionError):
        get_defined_fun(x1)
    with pytest.raises(AssertionError):
        get_defined_fun(x2)
    with pytest.raises(AssertionError):
        get_defined_fun(x3)
    with pytest.raises(AssertionError):
        get_defined_fun(fpadd)
    with pytest.raises(AssertionError):
        get_defined_fun(fpmul)
    with pytest.raises(AssertionError):
        get_defined_fun(fprem)
    with pytest.raises(AssertionError):
        get_defined_fun(fpfma)
    with pytest.raises(AssertionError):
        get_defined_fun(fadd)
    with pytest.raises(AssertionError):
        get_defined_fun(fmul)
    with pytest.raises(AssertionError):
        get_defined_fun(fx)
    with pytest.raises(AssertionError):
        get_defined_fun(gx)
    with pytest.raises(AssertionError):
        get_defined_fun(f)
    assert get_defined_fun(Node('fadd')) == fpadd
    assert get_defined_fun(Node('fmul')) == fpmul
    assert get_defined_fun(Node('fx')) == fprem
    assert get_defined_fun(Node('gx')) == fpfma
    assert get_defined_fun(Node('f')) == Node('+', 'x', 'x')
    assert get_defined_fun(Node('fx', 'true')) \
           == Node('fp.rem', 'true', 'true')
    assert get_defined_fun(Node('gx', 'true', 2)) \
           == Node('fp.fma', 2, 'fx', 'fx', 'fx')
    assert get_defined_fun(Node('f', 15)) == Node('+', 15, 15)
    assert get_defined_fun(Node('f', 15, 15)) == Node('f', 15, 15)


def test_is_set_sort():
    assert not is_set_sort(
        Node('declare-const', 'x', ('Set', ('_', 'BitVec', 8))))
    assert not is_set_sort(Node('Set'))
    assert not is_set_sort(Node('bvnot', 'x'))
    assert is_set_sort(Node('Set', 'Int'))
    assert is_set_sort(Node('Set', ('_', 'BitVec', 8)))



#
# ddSMT: A delta debugger for SMT benchmarks in SMT-Lib v2 format.
#
# This file is part of ddSMT.
#
# Copyright (C) 2013-2021 by the authors listed in AUTHORS file.
#
# ddSMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ddSMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ddSMT.  If not, see <https://www.gnu.org/licenses/>.

import collections
import multiprocessing
import struct


class Node:
    """Represents a node in the input, consisting of an id and some data.

    The ``data`` can either be a string or a tuple of nodes. The ``id``
    is automatically set to a unique integer that can be used for local
    substitutions.
    """
    __slots__ = 'id', 'data', 'hash'
    __ID_COUNTER = multiprocessing.Value('i', 0)

    @classmethod
    def __get_id(self):
        with self.__ID_COUNTER.get_lock():
            self.__ID_COUNTER.value += 1
            return self.__ID_COUNTER.value

    def __init__(self, *args, _id=None, _data=None, _hash=None):
        """
        Node("str") -> "str"
        otherwise:
        Node(*args) -> tuple(*args)
        If any of ``_id``, ``_data`` and ``_hash`` are set, these are used to
        populate the respective properties without checking them. Only use
        them if you really know what you are doing!
        """
        self.id = _id if _id else self.__get_id()
        if _data:
            self.data = _data
        else:
            if len(args) == 1 and isinstance(args[0], (str, int)):
                self.data = str(args[0])
            else:
                self.data = tuple(map(lambda a: self.__ensure_is_node(a),
                                      args))
        self.hash = _hash if _hash else hash(self.data)

    def __deepcopy__(self, memo):
        """Hook for copy.deepcopy, make sure we assign a fresh id."""
        visit = [(self, False)]
        args = [[]]
        while visit:
            expr, visited = visit.pop()
            assert isinstance(expr, Node)

            if visited:
                children = args.pop()
                node = Node(*children)
                args[-1].append(node)
            else:
                if expr.is_leaf():
                    args[-1].append(Node(expr.data))
                else:
                    visit.append((expr, True))
                    visit.extend((x, False) for x in reversed(expr.data))
                    args.append([])
        assert len(args) == 1
        if len(args[0]) == 0:
            return None
        return args[0][0]

    def __ensure_is_node(self, data):
        """Recursively walk data and make sure everything is a node."""
        if isinstance(data, Node):
            return data
        if isinstance(data, str):
            return Node(data)
        if isinstance(data, int):
            return Node(str(data))
        assert isinstance(data, tuple)
        return Node(*map(lambda a: self.__ensure_is_node(a), data))

    def __str__(self):
        if isinstance(self.data, str):
            return self.data
        return '(' + ' '.join(map(str, self.data)) + ')'

    def __repr__(self):
        if isinstance(self.data, str):
            return f'"{self.data}"'
        return f'({self.id} ' + ' '.join(map(repr, self.data)) + ')'

    def __len__(self):
        if self.is_leaf():
            return 0
        return len(self.data)

    def __getitem__(self, key):
        return self.data[key]

    def __eq__(self, other):
        if other is None:
            return False
        if isinstance(other, Node):
            if self.id == other.id:
                return True
            visit_other = [other]
        else:
            if isinstance(other, tuple):
                visit_other = [Node(*other)]
            elif isinstance(other, str):
                return self.is_leaf() and self.data == other
            else:
                visit_other = [Node(other)]
        visit_self = [self]
        while visit_self:
            if not visit_other:
                return False
            ns = visit_self.pop()
            no = visit_other.pop()
            assert isinstance(ns, Node)
            assert isinstance(no, Node)
            if ns.id == no.id:
                continue
            if ns.is_leaf() != no.is_leaf():
                return False
            if ns.hash != no.hash:
                return False
            if ns.is_leaf():
                if ns.data != no.data:
                    return False
            else:
                if len(ns) != len(no):
                    return False
                visit_self.extend(ns.data)
                visit_other.extend(no.data)
        return True

    def __hash__(self):
        return self.hash

    def __getstate__(self):
        """Callback method for custom (non-recursive) pickling."""
        res = []
        visit = [self]
        while visit:
            expr = visit.pop()
            if isinstance(expr, bytes):
                res.append(expr)
                continue
            assert isinstance(expr, Node)
            if expr.is_leaf():
                res.append(b'L')
                res.append(struct.pack("=ii", expr.id, len(expr.data)))
                res.append(expr.data.encode())
            else:
                res.append(b'(')
                res.append(struct.pack("=iq", expr.id, expr.hash))
                visit.append(b')')
                visit.extend(reversed(expr.data))

        return b''.join(res)

    def __setstate__(self, state):
        """Callback method for custom (non-recursive) unpickling."""
        exprs = [[]]
        i = 0
        smax = len(state)
        while i < smax:
            cur = state[i]
            if cur == 40:  # b'('
                exprs.append(list(struct.unpack('=iq', state[i + 1:i + 13])))
                i += 13
                continue
            if cur == 41:  # b')'
                i += 1
                children = exprs.pop()
                _id = children.pop(0)
                _hash = children.pop(0)
                node = Node(_data=tuple(children), _id=_id, _hash=_hash)
                exprs[-1].append(node)
                continue
            if cur == 76:  # b'L'
                _id, leaflen = struct.unpack('=ii', state[i + 1:i + 9])
                node = Node(state[i + 9:i + leaflen + 9].decode(), _id=_id)
                exprs[-1].append(node)
                i += leaflen + 9
                continue
            break
        self.id = exprs[0][0].id
        self.hash = exprs[0][0].hash
        self.data = exprs[0][0].data

    def is_leaf(self):
        """Return true if this node is a leaf node, i.e., it has no children
        but is only a string."""
        return isinstance(self.data, str)

    def has_ident(self):
        """Return true if this node has an identifier, i.e., it has children
        and its first child is a leaf node."""
        return (isinstance(self.data, tuple) and self.data
                and self.data[0].is_leaf())

    def get_ident(self):
        """Get the identifier of this, asserting that ``node.has_ident()``."""
        assert self.has_ident()
        return self.data[0]


def reduplicate(exprs):
    """Re-duplicates nodes with the same id using deepcopy."""
    ids = set()
    visit = [(x, False) for x in reversed(exprs)]
    args = [[]]
    while visit:
        expr, visited = visit.pop()
        assert isinstance(expr, Node)

        if expr.is_leaf():
            if expr.id in ids:
                args[-1].append(Node(expr.data))
            else:
                args[-1].append(expr)
                ids.add(expr.id)
        else:
            if visited:
                children = args.pop()
                if any(map(lambda x: x[0].id != x[1].id, zip(expr, children))):
                    node = Node(*children)
                    args[-1].append(node)
                    ids.add(node.id)
                else:
                    args[-1].append(expr)
                    ids.add(expr.id)
            else:
                visit.append((expr, True))
                visit.extend((x, False) for x in reversed(expr.data))
                args.append([])
    assert len(args) == 1
    return args[0]


def dfs(exprs, max_depth=None):
    """DFS traversal of s-expressions in exprs up to a maximum depth."""
    if isinstance(exprs, Node):
        yield exprs
    visit = [(1, x) for x in reversed(exprs)]
    while visit:
        cur_depth, expr = visit.pop()
        if isinstance(expr, Node) and (not max_depth or cur_depth < max_depth):
            yield expr
            if not expr.is_leaf():
                visit.extend([(cur_depth + 1, x) for x in reversed(expr.data)])
        else:
            yield expr


def bfs(exprs, max_depth=None):
    """BFS traversal of s-expressions in exprs up to a maximum depth."""
    if isinstance(exprs, Node):
        yield exprs
    visit = collections.deque()
    visit.extend([(1, x) for x in exprs])
    while visit:
        cur_depth, expr = visit.popleft()
        if isinstance(expr, Node) and (not max_depth or cur_depth < max_depth):
            yield expr
            if not expr.is_leaf():
                visit.extend([(cur_depth + 1, x) for x in expr.data])
        else:
            yield expr


def substitute(exprs, repl):  # noqa: C901
    """Performs (local and global) substitutions on exprs as specified in repl.
    repl may contain two types of entries:

    - int -> Node to replace nodes with the given id with the value node
    - Node -> Node to replace nodes equal to the key node with the value node
    """
    if len(repl) == 0:
        return exprs
    if isinstance(exprs, Node):
        if exprs.is_leaf():
            if exprs.id and exprs.id in repl:
                return repl[exprs.id]
            if exprs in repl:
                return repl[exprs]
            return exprs
        visit = [(exprs, False)]
    else:
        visit = [(x, False) for x in reversed(exprs)]
    changed = False
    args = [[]]
    while visit:
        expr, visited = visit.pop()
        assert isinstance(expr, Node)

        didrepl = False
        if expr.id and expr.id in repl:
            expr = repl.pop(expr.id)
            didrepl = True
        if expr in repl:
            expr = repl[expr]
            didrepl = True
        if didrepl:
            changed = True
            if expr is None:
                continue

        if visited:
            children = args.pop()
            node = Node(*children)
            if node == expr:
                args[-1].append(expr)
            else:
                args[-1].append(node)
        else:
            if not repl or expr.is_leaf():
                args[-1].append(expr)
            else:
                visit.append((expr, True))
                visit.extend((x, False) for x in reversed(expr.data))
                args.append([])
    if not changed:
        return exprs
    if isinstance(exprs, Node):
        assert len(args) == 1
        if len(args[0]) == 0:
            return None
        return args[0][0]
    return args[0]


def count_nodes(node):
    """Return the number of expressions yielded when traversing ``node`` in DFS
    manner."""
    assert isinstance(node, (Node, list))
    res = 0
    if isinstance(node, Node):
        visit = [node]
    else:
        visit = [x for x in node]
    while visit:
        expr = visit.pop()
        res += 1
        if not expr.is_leaf():
            visit.extend([x for x in expr.data])
    return res


def count_exprs(node):
    """Return the number of tuples yielded when traversing ``node`` in DFS
    manner."""
    assert isinstance(node, (Node, list))
    res = 0
    if isinstance(node, Node):
        visit = [node]
    else:
        visit = [x for x in node]
    while visit:
        expr = visit.pop()
        if not expr.is_leaf():
            res += 1
            visit.extend([x for x in expr.data if not x.is_leaf()])
    return res


def filter_nodes(exprs, filter_func, max_depth=-1):
    """Filter s-expressions based on filter_func."""
    assert isinstance(exprs, (Node, list))
    for expr in dfs(exprs, max_depth):
        if filter_func(expr):
            yield expr


def binary_search(input_length):
    """Assume an input sequence of some ``input_length``, generates indices of
    sub-sequences of decreasing length in a binary-search fashion."""
    den = 2
    while den * 2 <= input_length:
        for num in reversed(range(0, den)):
            start = int(num / den * input_length)
            end = int((num + 1) / den * input_length)
            yield (start, end)
        den *= 2


def contains(node, func):
    """Checks whether ``func`` returns true for any of the nodes within
    ``node``."""
    return any(map(func, dfs(node)))



import logging

if not hasattr(logging, 'trace'):
    logging.TRACE = 5
    logging.addLevelName(logging.TRACE, "TRACE")
    logging.trace = lambda msg, *args, **kwargs: logging.log(
        logging.TRACE, msg, *args, **kwargs)

if not hasattr(logging, 'chat'):
    logging.CHAT = 25
    logging.addLevelName(logging.CHAT, "CHAT")
    logging.chat = lambda msg, *args, **kwargs: logging.log(
        logging.CHAT, msg, *args, **kwargs)



#
# ddSMT: A delta debugger for SMT benchmarks in SMT-Lib v2 format.
#
# This file is part of ddSMT.
#
# Copyright (C) 2013-2021 by the authors listed in AUTHORS file.
#
# ddSMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ddSMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ddSMT.  If not, see <https://www.gnu.org/licenses/>.

import io
import textwrap
import typing

from .nodes import Node


def parse_smtlib(text: str):  # noqa: C901
    """Parse SMT-LIB input to list of ``Node`` objects.

    Every node represents an s-expression. This generator yields top-
    level s-expressions (commands) or comments.
    """
    exprs = []
    cur_expr = None

    pos = 0
    size = len(text)
    while pos < size:
        char = text[pos]
        pos += 1

        # String literals/quoted symbols
        if char in ('"', '|'):
            first_char = char
            literal = [char]
            # Read until terminating " or |
            while True:
                if pos >= size:
                    return
                char = text[pos]
                pos += 1
                literal.append(char)
                if char == first_char:
                    # Check is quote is escaped "a "" b" is one string literal
                    if char == '"' and pos < size and text[pos] == '"':
                        literal.append(text[pos])
                        pos += 1
                        continue
                    break
            cur_expr.append(Node(''.join(literal)))

        # Comments
        elif char == ';':
            comment = [char]
            # Read until newline
            while pos < size:
                char = text[pos]
                pos += 1
                comment.append(char)
                if char == '\n':
                    break
            comment = ''.join(comment)
            if cur_expr:
                cur_expr.append(Node(comment))
            else:
                yield Node(comment)

        # Open s-expression
        elif char == '(':
            cur_expr = []
            exprs.append(cur_expr)

        # Close s-expression
        elif char == ')':
            cur_expr = exprs.pop()

            # Do we have nested s-expressions?
            if exprs:
                exprs[-1].append(Node(*cur_expr))
                cur_expr = exprs[-1]
            else:
                yield Node(*cur_expr)
                cur_expr = None

        # Identifier
        elif char not in (' ', '\t', '\n'):
            token = [char]
            while True:
                if pos >= size:
                    return
                char = text[pos]
                pos += 1
                if char in (' ', '\t', '\n'):
                    break
                if char in ('(', ')', ';'):
                    pos -= 1
                    break
                token.append(char)

            token = Node(''.join(token))

            # Append to current s-expression
            if cur_expr is not None:
                cur_expr.append(token)
            else:
                yield token


def __write_smtlib(file: typing.TextIO, expr: Node):
    """Write the given smtlib expression in one line into the file object."""
    visit = [expr]
    needs_space = False
    while visit:
        ex = visit.pop()
        if ex is None:
            file.write(')')
            needs_space = True
            continue

        if needs_space:
            file.write(' ')

        if ex.is_leaf():
            if ex.data == '':
                continue
            if ex.data[0] == ';':
                file.write(f'\n{ex.data}\n')
            else:
                file.write(ex.data)
            needs_space = True
            continue

        file.write('(')
        needs_space = False
        visit.append(None)
        visit.extend(x for x in reversed(ex.data))


def __write_smtlib_pretty(file: typing.TextIO, expr: Node):
    """Write the given smtlib expression in one line into the file object."""
    visit = [(expr, False)]
    indent = ''
    while visit:
        ex, visited = visit.pop()
        if ex.is_leaf():
            assert isinstance(ex.data, str)
            if ex.data == '':
                continue
            if ex.data[0] == ';':
                file.write(f'\n{ex.data}\n')
            else:
                file.write(f'{indent}{ex.data}\n')
            continue

        if visited:
            indent = indent[:-2]
            file.write(f'{indent})\n')
        else:
            if all(map(lambda n: n.is_leaf(), ex.data)):
                file.write(f'{indent}{ex}\n')
            else:
                visit.append((ex, True))
                if ex.has_ident():
                    file.write(f'{indent}({ex.data[0]}\n')
                    visit.extend((x, False) for x in reversed(ex.data[1:]))
                else:
                    file.write(f'{indent}(\n')
                    visit.extend((x, False) for x in reversed(ex.data))
                indent += '  '


def __write_smtlib_str(expr: Node):
    """Write the given smtlib expression in one line to a string."""
    f = io.StringIO()
    __write_smtlib(f, expr)
    return f.getvalue()


def __write_smtlib_pretty_str(expr: Node):
    """Write the given smtlib expression to a formatted string."""
    f = io.StringIO()
    __write_smtlib_pretty(f, expr)
    return f.getvalue()


def write_smtlib(file: typing.TextIO, exprs: typing.List[Node]):
    """Write the given expressions to the given file object
    Honor options to wrap lines or pretty-print."""
    from . import options
    if options.args().pretty_print:
        # pretty print
        for expr in exprs:
            __write_smtlib_pretty(file, expr)
    else:
        # regular writeing
        lines = [__write_smtlib_str(expr) for expr in exprs]
        if options.args().wrap_lines:
            # wrap every line
            lines = map(
                lambda line: textwrap.wrap(
                    line, width=78, subsequent_indent='  '), lines)
            # and flatten the list
            lines = [sub for line in lines for sub in line]
        for line in lines:
            file.write(line)
            file.write('\n')


def write_smtlib_to_file(filename: str, exprs: typing.List[Node]):
    """Use ``write_smtlib`` to write to a filename."""
    with open(filename, 'w') as file:
        write_smtlib(file, exprs)


def write_smtlib_to_str(exprs: typing.List[Node]):
    """Return result of ``write_smtlib`` as a string."""
    f = io.StringIO()
    write_smtlib(f, exprs)
    return f.getvalue()


def write_smtlib_for_checking(filename: str, exprs: typing.List[Node]):
    """Slightly faster writing without wrapping or pretty-printing during
    checking."""
    with open(filename, 'w') as file:
        for expr in exprs:
            __write_smtlib(file, expr)



from ..nodes import Node
from .. import mutators_fp
from .utils import *


def test_fp_get_mutators():
    d = mutators_fp.get_mutators()
    assert isinstance(d, dict)
    assert len(d) == 1


def test_fp_is_relevant():
    assert mutators_fp.is_relevant(Node('declare-const', 'x', 'Float16'))
    assert mutators_fp.is_relevant(
        Node('declare-fun', 'x', (), ('_', 'FloatingPoint', '5', '11')))
    assert not mutators_fp.is_relevant(Node('declare-const', 'x', 'Real'))
    assert not mutators_fp.is_relevant(Node('declare-fun', 'x', (), 'Bool'))
    assert not mutators_fp.is_relevant(Node())
    assert not mutators_fp.is_relevant(Node('assert', 'x'))


def test_fp_short_sort():
    m = mutators_fp.FPShortSort()
    assert isinstance(str(m), str)
    assert not m.filter(Node('x'))
    assert not m.filter(
        Node('declare-const', 'x', ('_', 'FloatingPoint', 3, 5)))
    assert not m.filter(Node('Float16'))
    assert not m.filter(Node('_', 'FloatingPoint', 5))
    assert not m.filter(Node('_', 'FloatingPoin', 5, 11))
    assert m.filter(Node('_', 'FloatingPoint', 3, 5))
    assert check_mutations(m, Node('_', 'FloatingPoint', 3, 5), [])
    assert check_mutations(m, Node('_', 'FloatingPoint', 5, 11),
                           [Node('Float16')])
    assert check_mutations(m, Node('_', 'FloatingPoint', 8, 24),
                           [Node('Float32')])
    assert check_mutations(m, Node('_', 'FloatingPoint', 11, 53),
                           [Node('Float64')])
    assert check_mutations(m, Node('_', 'FloatingPoint', 15, 113),
                           [Node('Float128')])



#
# ddSMT: A delta debugger for SMT benchmarks in SMT-Lib v2 format.
#
# This file is part of ddSMT.
#
# Copyright (C) 2013-2021 by the authors listed in AUTHORS file.
#
# ddSMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ddSMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ddSMT.  If not, see <https://www.gnu.org/licenses/>.

import collections
import logging
import multiprocessing
import pickle
import sys
import time
import traceback

from . import checker
from . import mutators
from . import nodeio
from . import nodes
from . import options
from . import debug_utils
from . import progress
from . import smtlib
from .mutator_utils import Simplification, apply_simp


def get_passes():
    """Returns a list of passes, each pass being a list of mutators.

    The list is ordered so that earlier passes are more promising for a
    quick reduction.
    """
    prelude = [
        (mutators.get_initialized_mutator('BinaryReduction',
                                          {'ident': 'assert'}), {
                                              'max_depth': 1
                                          }),
        mutators.get_mutators(['Constants', 'BinaryReduction']),
        mutators.get_mutators([
            'EraseNode',
            'Constants',
            'IntroduceFreshVariable',
            'LetElimination',
            'LetSubstitution',
            'ReplaceByChild',
            'BinaryReduction',
            'CheckSatAssuming',
        ]),
    ]

    late = [  # Usually only have cosmetic impact
        'SimplifyQuotedSymbols',
        'SimplifySymbolNames',
        'StringSimplifyConstant',
    ]
    main = []
    for _, theory in mutators.get_all_mutators().items():
        for mname in theory[1]:
            if mname not in late:
                main.append(mname)

    return prelude + [
        mutators.get_mutators(main),
        mutators.get_mutators(late + main),
    ]


def get_pass(passes, id):
    """Return the current pass, add empty parameter dict if none is given
    explicitly."""
    res = passes[id]
    if isinstance(res, tuple):
        return res
    return res, {}


# nodeid: id of the mutated node in bfs order. Only used for progress indication
# name: name of the mutator
# exprs: the current input
# simp: the substitution to be checked
# runtime: time needed to check this task
Task = collections.namedtuple('Task',
                              ['nodeid', 'name', 'exprs', 'simp', 'runtime'])


class Producer:
    """Manages the generation of candidates that shall be checked.

    Performs a walk through the current input and applies the
    ``mutators`` to every node. Supports skipping the first ``skip``
    nodes. As soon as ``abort_flag`` is triggered, stops generation as
    soon as possible.
    """
    def __init__(self, mutators, abort_flag, original):
        self.__node_count = 0
        self.__mutators = mutators
        self.__abort = abort_flag
        self.__original = original
        self.__pickled = pickle.dumps(original)

    def __mutate_node(self, count, linput):
        """Apply all mutators to the given node.

        Returns a list of all possible mutations as ``Task`` objects.
        """
        for m in self.__mutators:
            if self.__abort.is_set():
                break
            try:
                if hasattr(m, 'filter') and not m.filter(linput):
                    continue
                if hasattr(m, 'mutations'):
                    for x in m.mutations(linput):
                        if self.__abort.is_set():
                            break
                        assert isinstance(x, Simplification)
                        yield Task(count, str(m), self.__pickled,
                                   pickle.dumps(x), None)
                if hasattr(m, 'global_mutations'):
                    for x in m.global_mutations(linput, self.__original):
                        if self.__abort.is_set():
                            break
                        assert isinstance(x, Simplification)
                        yield Task(count, f'(global) {m}', self.__pickled,
                                   pickle.dumps(x), None)
            except Exception as e:
                logging.info(f'{type(e)} in application of {m}: {e}')
                exc_type, exc_value, exc_traceback = sys.exc_info()
                traceback.print_tb(exc_traceback, limit=10, file=sys.stderr)

    def generate(self, skip, params):
        """A generator that produces all possible mutations as ``Task`` from
        the given original."""
        count = 0
        for node in nodes.bfs(self.__original, params.get('max_depth', None)):
            count += 1
            if skip < count:
                yield from self.__mutate_node(count, node)
            if self.__abort.is_set():
                break


class Consumer:
    """Calls the ``checker`` on individual tasks to figure out whether they are
    valid simplifications.

    Uses the ``abort_flag`` to stop as soon as a valid simplification
    has been found.
    """
    def __init__(self, abort_flag):
        self.__abort = abort_flag

    def check(self, task):
        with debug_utils.Profiler():
            abortres = pickle.dumps(
                (False, Task(task.nodeid, task.name, None, None, None)))
            if self.__abort.is_set():
                return abortres
            try:
                start = time.time()
                simp = pickle.loads(task.simp)
                assert isinstance(simp, Simplification)
                if self.__abort.is_set():
                    return abortres
                exprs = apply_simp(pickle.loads(task.exprs), simp)

                if self.__abort.is_set():
                    return abortres
                res = checker.check_exprs(exprs)
                runtime = time.time() - start
                if self.__abort.is_set():
                    return abortres
                if res:
                    return pickle.dumps((True,
                                         Task(task.nodeid, task.name, exprs,
                                              None, runtime)))
                return pickle.dumps(
                    (False, Task(task.nodeid, task.name, None, None, runtime)))
            except Exception as e:
                logging.info(f'{type(e)} in check of {task.name}: {e}')
                exc_type, exc_value, exc_traceback = sys.exc_info()
                traceback.print_tb(exc_traceback, limit=10, file=sys.stderr)
            return abortres


class MutatorStats:
    """Gather information about the performance of the individual mutators."""
    def __init__(self):
        self.data = {}
        self.__enabled = logging.getLogger().isEnabledFor(logging.INFO)

    def add(self, success, task, original):
        """Add results from one check."""
        if task.runtime is None or not self.__enabled:
            return
        d = self.data.setdefault(task.name, {
            'tests': 0,
            'success': 0,
            'diff': 0,
            'runtime': 0
        })
        d['tests'] += 1
        if success:
            d['success'] += 1
            d['diff'] += nodes.count_exprs(
                task.exprs) - nodes.count_exprs(original)
        d['runtime'] += task.runtime

    def print(self):
        """Print cumulative results."""
        if self.__enabled:
            for name, data in sorted(self.data.items()):
                logging.info(f'{name}: diff {data["diff"]:+} expressions, '
                             f'{data["tests"]} tests ({data["success"]}), '
                             f'{data["runtime"]:.1f}s')


def reduce(exprs):
    """Reduces the input given in ``exprs`` as good as possible in a fixed-
    point loop."""
    passes = get_passes()
    passid = 0

    nchecks = 0
    nreduce = 0
    stats = MutatorStats()
    loop_checker = debug_utils.NodeLoopChecker()
    loop_checker.add(exprs)

    # use one pool for the whole reduction
    with multiprocessing.Pool(options.args().jobs) as pool:
        # abort flag is passed to both producer and consumer
        # important: the pool will break if the abort_flag is destroyed early
        abort_flag = multiprocessing.Manager().Event()
        # iterate over all passes
        for passid in range(len(passes)):
            cur_passes, params = get_pass(passes, passid)
            logging.chat(f'pass {passid + 1} / {len(passes)}')
            skip = 0
            fresh_run = True
            # repeat until no further reduction is found
            while True:
                reduction = False
                start = time.time()
                smtlib.collect_information(exprs)
                cnt = nodes.count_nodes(exprs)
                progress.start(cnt)
                progress.update(min(cnt, skip))
                abort_flag.clear()
                prod = Producer(cur_passes, abort_flag, exprs)
                cons = Consumer(abort_flag)
                for result in pool.imap_unordered(cons.check,
                                                  prod.generate(skip, params)):
                    nchecks += 1
                    success, task = pickle.loads(result)
                    if abort_flag.is_set():
                        # skip remaining results if we had a success
                        skip = min(skip, task.nodeid - 1)
                        continue
                    progress.update(task.nodeid)
                    stats.add(success, task, exprs)
                    if success:
                        # trigger abort, then process the result
                        abort_flag.set()
                        progress.finish()
                        nreduce += 1
                        runtime = time.time() - start
                        logging.chat(
                            f'#{nreduce}: {task.name} ({runtime:.2f}s, '
                            f'{nodes.count_nodes(task.exprs)} expressions)')
                        reduction = True
                        debug_utils.dump_diff(task.name, exprs, task.exprs)
                        exprs = nodes.reduplicate(task.exprs)
                        loop_checker.add(exprs)
                        skip = task.nodeid - 1
                        fresh_run = False
                        nodeio.write_smtlib_to_file(options.args().outfile,
                                                    exprs)
                if not reduction:
                    progress.finish()
                    logging.info('No further simplification found')
                    if fresh_run:
                        # this was a fresh run, continue with next pass
                        break
                    logging.info('Starting over')
                    skip = 0
                    fresh_run = True

    stats.print()

    return exprs, nchecks



#
# ddSMT: A delta debugger for SMT benchmarks in SMT-Lib v2 format.
#
# This file is part of ddSMT.
#
# Copyright (C) 2013-2021 by the authors listed in AUTHORS file.
#
# ddSMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ddSMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ddSMT.  If not, see <https://www.gnu.org/licenses/>.

from . import nodes
from .nodes import Node
from .smtlib import *
from .mutator_utils import Simplification


class FPShortSort:
    """Replace long sort names with short names.

    .. code-block:: common-lisp

        (_ FloatingPoint 5 11)   --> Float16
        (_ FloatingPoint 8 24)   --> Float32
        (_ FloatingPoint 11 53)  --> Float64
        (_ FloatingPoint 15 113) --> Float128
    """
    def filter(self, node):
        return is_fp_sort(node) and len(node) == 4

    def mutations(self, node):
        if node[2] == '5' and node[3] == '11':
            return [Simplification({node.id: Node('Float16')}, [])]
        if node[2] == '8' and node[3] == '24':
            return [Simplification({node.id: Node('Float32')}, [])]
        if node[2] == '11' and node[3] == '53':
            return [Simplification({node.id: Node('Float64')}, [])]
        if node[2] == '15' and node[3] == '113':
            return [Simplification({node.id: Node('Float128')}, [])]
        return []

    def __str__(self):
        return 'replace long FP sort name with short name'


def get_mutators():
    """Return mapping from mutator class names to the name of their config
    options."""
    return {
        'FPShortSort': 'fp-short-sort',
    }


def is_relevant(node):
    """Checks whether this theory might be relevant for this node."""
    if node.has_ident():
        if node.get_ident() in ['declare-const']:
            if nodes.contains(node[2], is_fp_sort):
                return True
        elif node.get_ident() in ['declare-fun', 'define-fun', 'define-sort']:
            if nodes.contains(node[3], is_fp_sort):
                return True
    return False



from ..nodes import Node
from .. import mutators_boolean
from .utils import *


def test_bool_get_mutators():
    d = mutators_boolean.get_mutators()
    assert isinstance(d, dict)
    assert len(d) == 7


def test_bool_de_morgan():
    m = mutators_boolean.BoolDeMorgan()
    assert isinstance(str(m), str)
    notfilter = [
        Node('=', '1', 'y'),
        Node('and', 'x', 'y'),
        Node('or', 'x', 'y'),
    ]
    exprs = {
        Node('not', ('and', 'x', 'y', 'z')):
        [Node('or', ('not', 'x'), ('not', 'y'), ('not', 'z'))],
        Node('not', ('or', 'x', 'y')):
        [Node('and', ('not', 'x'), ('not', 'y'))],
    }
    for e in notfilter:
        assert not m.filter(e)
    for k, v in exprs.items():
        assert m.filter(k)
        assert check_mutations(m, k, v)


def test_bool_double_negation():
    m = mutators_boolean.BoolDoubleNegation()
    assert isinstance(str(m), str)
    notfilter = [
        Node('not', ('=', '1', 'y')),
        Node('not', ('and', 'x', 'y')),
        Node('not', ('or', 'x', 'y')),
    ]
    exprs = {
        Node('not', ('not', 'x')): [Node('x')],
        Node('not', ('not', ('or', 'x', 'y'))): [Node('or', 'x', 'y')],
    }
    for e in notfilter:
        assert not m.filter(e)
    for k, v in exprs.items():
        assert m.filter(k)
        assert check_mutations(m, k, v)


def test_bool_eliminate_false_equality():
    m = mutators_boolean.BoolEliminateFalseEquality()
    assert isinstance(str(m), str)
    notfilter = [
        Node('not', ('=', '1', 'y')),
        Node('=', 'true', 'x', 'y'),
        Node('not', ('=', 'false', 'y')),
    ]
    exprs = {
        Node('=', 'false'): [],
        Node('=', 'false', 'x'): [Node('not', 'x')],
        Node('=', 'x', 'false', 'y'):
        [Node('and', ('not', 'x'), ('not', 'y'))],
    }
    for e in notfilter:
        assert not m.filter(e)
    for k, v in exprs.items():
        assert m.filter(k)
        assert check_mutations(m, k, v)


def test_bool_eliminate_implication():
    m = mutators_boolean.BoolEliminateImplication()
    assert isinstance(str(m), str)
    notfilter = [
        Node('or', 'true', 'y'),
        Node('not', ('=>', 'false', 'y')),
    ]
    exprs = {
        Node('=>', 'x'): [],
        Node('=>', 'x', 'y'): [Node('or', ('not', 'x'), 'y')],
        Node('=>', 'x', 'y', 'z'):
        [Node('and', ('or', ('not', 'x'), 'y'), ('or', ('not', 'y'), 'z'))],
    }
    for e in notfilter:
        assert not m.filter(e)
    for k, v in exprs.items():
        assert m.filter(k)
        assert check_mutations(m, k, v)


def test_bool_negate_quantifier():
    m = mutators_boolean.BoolNegateQuantifier()
    assert isinstance(str(m), str)
    notfilter = [
        Node('forall', ('Real', 'x'), 'y'),
        Node('not', ('or', 'x', 'y')),
    ]
    exprs = {
        Node('not', ('forall', ('Real', 'x'), 'y')):
        [Node('exists', ('Real', 'x'), ('not', 'y'))],
        Node('not', ('exists', ('Real', 'x'), 'y')):
        [Node('forall', ('Real', 'x'), ('not', 'y'))],
    }
    for e in notfilter:
        assert not m.filter(e)
    for k, v in exprs.items():
        assert m.filter(k)
        assert check_mutations(m, k, v)


def test_bool_xor_eliminate_binary():
    m = mutators_boolean.BoolXOREliminateBinary()
    assert isinstance(str(m), str)
    notfilter = [
        Node('or', 'x', 'y'),
        Node('xor', 'x', 'y', 'z'),
    ]
    exprs = {
        Node('xor', 'x', 'y'): [Node('distinct', 'x', 'y')],
    }
    for e in notfilter:
        assert not m.filter(e)
    for k, v in exprs.items():
        assert m.filter(k)
        assert check_mutations(m, k, v)


def test_bool_xor_remove_constant():
    m = mutators_boolean.BoolXORRemoveConstant()
    assert isinstance(str(m), str)
    notfilter = [
        Node('or', 'x', 'y', 'false'),
        Node('and', 'true', 'z'),
    ]
    exprs = {
        Node('xor', 'x', 'y'): [],
        Node('xor', 'false', 'y'): [Node('xor', 'y')],
        Node('xor', 'false', 'y', 'z', 'false'): [Node('xor', 'y', 'z')],
        Node('xor', 'true', 'y'):
        [Node('xor', 'y'), Node('not', ('xor', 'y'))],
        Node('xor', 'true', 'y', 'z', 'true'):
        [Node('xor', 'y', 'z'),
         Node('not', ('xor', 'y', 'z'))],
    }
    for e in notfilter:
        assert not m.filter(e)
    for k, v in exprs.items():
        assert m.filter(k)
        assert check_mutations(m, k, v)



#
# ddSMT: A delta debugger for SMT benchmarks in SMT-Lib v2 format.
#
# This file is part of ddSMT.
#
# Copyright (C) 2013-2021 by the authors listed in AUTHORS file.
#
# ddSMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ddSMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ddSMT.  If not, see <https://www.gnu.org/licenses/>.

import logging
import progressbar
import sys

__BAR = None


def start(max):
    """Initialize a new progress bar for at most ``max`` steps.

    Only initialize if the current log level is at least
    ``logging.INFO``.
    """
    global __BAR
    if not sys.stdout.isatty():
        return
    if logging.getLogger().level > logging.INFO:
        return
    widgets = [progressbar.Bar(), ' ', progressbar.Counter(), ' / ', str(max)]
    __BAR = progressbar.ProgressBar(maxval=max, widgets=widgets)
    __BAR.start()
    __BAR.update_interval = 1


def update(newval=None):
    """Update the current progress bar.

    The value is incremented by one, or set to newval if newval is not
    ``None``.
    """
    global __BAR
    if __BAR:
        if newval is not None:
            __BAR.update(min(newval, __BAR.maxval))
        else:
            __BAR.update(__BAR.currval + 1)


def finish():
    """Stop the current progress bar.

    Delete the object and write a newline.
    """
    global __BAR
    if __BAR:
        __BAR = None
        sys.stdout.write('\n')



#
# ddSMT: A delta debugger for SMT benchmarks in SMT-Lib v2 format.
#
# This file is part of ddSMT.
#
# Copyright (C) 2013-2021 by the authors listed in AUTHORS file.
#
# ddSMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ddSMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ddSMT.  If not, see <https://www.gnu.org/licenses/>.

import collections
from . import smtlib
from . import nodes

Simplification = collections.namedtuple('Simplification',
                                        ['substs', 'fresh_vars'])


def apply_simp(exprs, simp):
    assert isinstance(simp, Simplification)
    mexprs = nodes.substitute(exprs, simp.substs)
    if mexprs is not exprs and simp.fresh_vars:
        assert isinstance(mexprs, list)
        mexprs = smtlib.introduce_variables(mexprs, simp.fresh_vars)
    return mexprs



#
# ddSMT: A delta debugger for SMT benchmarks in SMT-Lib v2 format.
#
# This file is part of ddSMT.
#
# Copyright (C) 2013-2021 by the authors listed in AUTHORS file.
#
# ddSMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ddSMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ddSMT.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import os

from . import argparsemod
from . import version


class ToggleAction(argparse.Action):
    '''A simple ``argparse.Action`` class that is used for option pairs of the
    form ``--option`` and ``--no-option``.'''
    def __init__(self, opt_name, default=True, dest=None, help=None):
        super(ToggleAction,
              self).__init__([f'--{opt_name}', f'--no-{opt_name}'],
                             dest=dest,
                             nargs=0,
                             help=help,
                             default=default)

    def _get_value(self, option_string):
        '''Figure out whether the option shall be true or false.'''
        return not option_string.startswith('--no-')

    def __call__(self, parser, namespace, values, option_string=None):
        '''Set the option depending on which option was used.'''
        setattr(namespace, self.dest, self._get_value(option_string))


class CustomFormatter(argparse.ArgumentDefaultsHelpFormatter,
                      argparse.HelpFormatter):
    '''A custom formatter for printing the commandline help.

    Uses ``argparse.ArgumentDefaultsHelpFormatter`` to print default
    values, but avoids printing for ``ToggleAction`` options and
    ``None`` defaults. Furthermore uses the ``argparse.HelpFormatter``,
    to slightly increase the width reserved for the options.
    '''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, max_help_position=35)

    def _get_help_string(self, action):
        '''Render the help string.

        Appends the default, if needed.
        '''
        if action.default is None or isinstance(action, ToggleAction):
            return action.help
        return super(CustomFormatter, self)._get_help_string(action)

    def _format_action_invocation(self, action):
        '''Formats the options for the help page, uses a special formatting for
        ``ToggleAction`` options.'''
        if isinstance(action, ToggleAction):
            name = action.option_strings[0][2:]
            return ', '.join([f'--[no-]{name}'] + action.option_strings[2:])
        return super(CustomFormatter, self)._format_action_invocation(action)


class DumpConfigAction(argparse.Action):
    '''Dump the current config using ``pprint``.'''
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, True)
        import pprint
        pprint.pprint(vars(namespace))


def parse_options(mutators, cmdlineoptions=None):
    '''Configures the commandline parse and then parse the commandline options.

    ``cmdlineoptions`` can be used to override the contents of
    ``sys.argv``, but is mostly useful for unit tests.
    '''

    usage = 'ddsmt [<options>] <infile> <outfile> <cmd> [<cmd options>]'
    ap = argparsemod.ModularArgumentParser(
        usage=usage,
        formatter_class=CustomFormatter,
        modular_help_groups=['mutator help'])
    ap.add_argument('infile', help='the input file (in SMT-LIB v2 format)')
    ap.add_argument('outfile', help='the output file')
    ap.add_argument('cmd',
                    nargs=argparse.REMAINDER,
                    help='the command (with optional arguments)')

    ap.add_argument('--version', action='version', version=version.VERSION)
    ap.add_argument('-v',
                    action='count',
                    dest='verbosity',
                    default=0,
                    help='increase verbosity')
    ap.add_argument('-q',
                    action='count',
                    dest='quietness',
                    default=0,
                    help='decrease verbosity')
    ap.add_argument('--dump-config',
                    nargs=0,
                    action=DumpConfigAction,
                    help='dump configuration')
    ap.add_argument('--parser-test',
                    action='store_true',
                    help='only test the parser')
    ap.add_argument('--pretty-print',
                    action='store_true',
                    default=False,
                    help='pretty-print to output file')
    ap.add_argument('--wrap-lines',
                    action='store_true',
                    default=False,
                    help='wrap lines in output file')
    ap.add_argument('--strategy',
                    choices=['ddmin', 'hierarchical', 'hybrid'],
                    default='hybrid',
                    help='minimization strategy')

    apcheck = ap.add_argument_group('checker arguments')
    apcheck.add_argument('--unchecked',
                         action='store_true',
                         help='assume every change is okay without checking')
    apcheck.add_argument('-j',
                         '--jobs',
                         type=int,
                         metavar='n',
                         default=os.cpu_count() - 2,
                         help='number of parallel checks')
    apcheck.add_argument('--memout',
                         type=int,
                         metavar='megabytes',
                         help='memout for individual checks')
    apcheck.add_argument('--timeout',
                         metavar='timeout',
                         type=float,
                         help='timeout for test runs in seconds '
                         '(default: 1.5 * golden runtime)')
    apcheck.add_argument(
        '--ignore-output',
        action='store_true',
        help='ignore stdout and stderr, only consider exit code')
    apcheck.add_argument(
        '--match-err',
        metavar='str',
        help='match string in stderr to identify failing input')
    apcheck.add_argument(
        '--match-out',
        metavar='str',
        help='match string in stdout to identify failing input')

    apcheck.add_argument('-c',
                         '--cross-check',
                         metavar='cmd-cc',
                         dest='cmd_cc',
                         help='cross check command')
    apcheck.add_argument(
        '--timeout-cc',
        metavar='timeout',
        type=float,
        help='timeout for test runs of the cross check in seconds '
        '(default: 1.5 * golden runtime)')
    apcheck.add_argument(
        '--ignore-output-cc',
        action='store_true',
        help='ignore stdout and stderr, only consider exit code for cross check'
    )
    apcheck.add_argument(
        '--match-err-cc',
        metavar='str',
        help='match string in stderr to identify failing input for cross check'
    )
    apcheck.add_argument(
        '--match-out-cc',
        metavar='str',
        help='match string in stdout to identify failing input for cross check'
    )

    apdebug = ap.add_argument_group('debug arguments')
    apdebug.add_argument('--check-loops',
                         action='store_true',
                         default=False,
                         help='check for loops in the minimization process')
    apdebug.add_argument('--profile',
                         action='store_true',
                         default=False,
                         help='use cProfile for profiling')
    apdebug.add_argument('--dump-diffs',
                         action='store_true',
                         default=False,
                         help='dump a diff of every simplification')

    mutators.collect_mutator_options(ap)

    res = ap.parse_args(cmdlineoptions)

    if res.cmd_cc:
        res.cmd_cc = res.cmd_cc.split()

    return res


__PARSED_ARGS = None


def args(cmdlineoptions=None):
    '''Returns the commandline options.

    Calls ``parse_options()`` if parsing has not yet happened. The
    ``cmdlineoptions`` is passed on to ``parse_options`` in this case.
    '''
    global __PARSED_ARGS
    if __PARSED_ARGS is None:
        # make sure that options as a whole does not depend on other modules
        from . import mutators
        __PARSED_ARGS = parse_options(mutators, cmdlineoptions)
    return __PARSED_ARGS
